# (C) Copyright NuoDB, Inc. 2017-2020
#
# This source code is licensed under the MIT license found in the LICENSE
# file in the root directory of this source tree.

import logging
import os
import re
import requests
import threading

from pynuoca.nuoca_plugin import NuocaMPInputPlugin
from pynuoca.nuoca_util import nuoca_log, nuoca_gettimestamp, IntervalSync
from requests.auth import HTTPBasicAuth

# Unverified HTTPS request is being made because the NuoDB Admin service does
# not yet support service certificate verification.
from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

# NuoAdminMonitor plugin
#
# Example NuoAdminMonitor plugin configuration for Legacy Admin Layer
#
# - NuoAdminMonitor:
#    description : Monitor NuoDB Admin layer
#    admin_host: localhost
#    admin_rest_api_port: 8888
#    domain_username: domain
#    domain_password: bird
#    admin_collect_interval: 10
#    admin_collect_timeout: 1
#
# Example NuoAdminMonitor plugin configuration for New Admin Layer
#
# - NuoAdminMonitor:
#    description : Monitor NuoDB Admin layer
#    nuocaCollectionName: NuoAdminMon2
#    api_server: localhost:8888
#    client_key: /opt/nuodb/tls-config/keys/nuocmd.pem
#    server_cert: None
#    admin_collect_interval: 10
#    admin_collect_timeout: 1


class NuoAdminMonitorPlugin(NuocaMPInputPlugin):

  # NOTE: member variables that start with "_legacyAdmin_" refer to
  # the legacy Admin layer.  Member variables that start with
  # "_newAdmin_" refer to New (version 3.3) Admin layer

  DEFAULT_API_SERVER = 'localhost:8888'

  def __init__(self, parent_pipe):
    super(NuoAdminMonitorPlugin, self).__init__(parent_pipe, 'NuoAdminMon')
    self._config = None
    self._enabled = False
    self._database_regex_pattern = '.*'
    self._is_legacyAdmin = None
    self._nuocaCollectionName = 'NuoAdminMon'  # default

    #
    # Legacy Admin
    self._legacyAdmin_host = None
    self._legacyAdmin_rest_api_port = 8888
    self._legacyAdmin_auth = None
    self._legacyAdmin_base_url = None
    #
    # API endpoints:
    self._legacyAdmin_domain_enforcer_url = None
    #
    # self._numon_handler_ready = False
    self._legacyAdmin_domain_username = 'domain'
    self._legacyAdmin_domain_password = 'bird'
    self._domain_metrics = None
    self._host_uuid_shortname = False
    self._timer_thrd = None
    #
    # Variables commmon to _legacyAdmin_ and _new_
    self._admin_collect_interval = 10
    self._admin_collect_timeout = 1
    self._admin_collect_sync = None
    self._monitor_collect_queue = []
    #
    # New Admin
    self._newAdmin_api_server = NuoAdminMonitorPlugin.DEFAULT_API_SERVER
    self._newAdmin_client_key = None
    self._newAdmin_server_cert = None
    self._newAdmin_base_url = None
    self._newAdmin_requests_server_verify = False

    self._newAdmin_regions_url = None
    self._newAdmin_archives_url = None
    self._newAdmin_databases_url = None
    self._newAdmin_kvstore_url = None
    self._newAdmin_peers_url = None
    self._newAdmin_processes_url = None

  def _get_newAdmin_url(self):
    base_url = self._newAdmin_api_server
    if (not base_url.startswith('http://') and
          not base_url.startswith('https://')):  # noqa
      if self._newAdmin_client_key is None and \
              self._newAdmin_server_cert is None:
        base_url = 'http://' + base_url
      else:
        base_url = 'https://' + base_url
    return base_url

  @property
  def monitor_collect_queue(self):
    return self._monitor_collect_queue

  def get_legacyAdmin_rest_url(self, rest_url):
    try:
      req = requests.get(rest_url, auth=self._legacyAdmin_auth,
                         timeout=self._admin_collect_timeout)
      if req.status_code != 200:
        err_msg = "NuoAdminMon: Error code '%d' when " \
                  "calling Admin Rest API: %s" \
                  % (req.status_code, rest_url)
        nuoca_log(logging.ERROR, err_msg)
        return {'nuoca_collection_error': err_msg}
      return req.json()
    except requests.RequestException as e:
      err_msg = str(e)
      nuoca_log(logging.ERROR, err_msg)
      return {'nuoca_collection_error': err_msg}
    except Exception as e:
      err_msg = str(e)
      nuoca_log(logging.ERROR, err_msg)
      return {'nuoca_collection_error': err_msg}

  def get_newAdmin_rest_url(self, rest_url):
    get_data = True
    rval = {'data': []}
    offset = 0
    limit = 20
    try:
      while get_data:
        rest_url_page = "%s?offset=%d&limit=%d" % (rest_url, offset, limit)
        req = requests.get(rest_url_page, cert=self._newAdmin_client_key,
                           verify=self._newAdmin_requests_server_verify,
                           timeout=self._admin_collect_timeout)
        if req.status_code != 200:
          err_msg = "NuoAdminMon: Error code '%d' when " \
                    "calling Admin Rest API: %s" \
                    % (req.status_code, rest_url)
          nuoca_log(logging.ERROR, err_msg)
          return {'nuoca_collection_error': err_msg}
        req_response = req.json()
        if 'hasNext' in req_response and req_response['hasNext']:
          get_data = True
          offset += limit
        else:
          get_data = False
        if 'data' in req_response:
          for item in req_response['data']:
            rval['data'].append(item)
        else:
          rval['data'].append(req_response)

      return rval
    except requests.RequestException as e:
      err_msg = str(e)
      nuoca_log(logging.ERROR, err_msg)
      return {'nuoca_collection_error': err_msg}
    except Exception as e:
      err_msg = str(e)
      nuoca_log(logging.ERROR, err_msg)
      return {'nuoca_collection_error': err_msg}

  def get_legacyAdmin_enforcer(self):
    return self.get_legacyAdmin_rest_url(self._legacyAdmin_domain_enforcer_url)

  def get_legacyAdmin_regions(self):
    return self.get_legacyAdmin_rest_url(self._newAdmin_regions_url)

  def get_newAdmin_regions(self):
    return self.get_newAdmin_rest_url(self._newAdmin_regions_url)

  def get_newAdmin_archives(self):
    return self.get_newAdmin_rest_url(self._newAdmin_archives_url)

  def get_newAdmin_databases(self):
    return self.get_newAdmin_rest_url(self._newAdmin_databases_url)

  def get_newAdmin_database(self, dbname):
    db_url = '%s/%s' % (self._newAdmin_databases_url, dbname)
    return self.get_newAdmin_rest_url(db_url)

  def get_newAdmin_kvstore(self):
    return self.get_newAdmin_rest_url(self._newAdmin_kvstore_url)

  def get_newAdmin_peers(self):
    return self.get_newAdmin_rest_url(self._newAdmin_peers_url)

  def get_newAdmin_processes(self):
    return self.get_newAdmin_rest_url(self._newAdmin_processes_url)

  def _legacyAdmin_collector_thread(self, collect_timestamp):
    region_result_desired_fields = [u'region']
    host_result_desired_fields = \
      [u'stableId', u'hostname', u'port', u'version', u'address',
       u'isBroker', u'ipaddress', u'id']
    host_tags_desired_fields = \
      [u'default_archive_base', u'journal_base', u'archive_base',
       u'default_region', u'region', u'os_num_cpu', u'cores', u'os_num_cores',
       u'default_journal_base', u'ostype', u'cputype', u'os_ram_mb',
       u'osversion', u'os_num_fs']
    processes_desired_fields = \
      [u'status', u'uid', u'hostname', u'pid', u'nodeId', u'port',
       u'version', u'address', u'agentid', u'type', u'dbname']
    databases_desired_fields = \
    [u'status', u'name', u'tagConstraints', u'variables', u'groupOptions',
     u'archivesByGroup', u'archives', u'template', u'active',
     u'unmet_messages', u'options', u'ismet']
    database_boolean_fields = \
    [u'active', u'ismet']

    collect_timestamp = nuoca_gettimestamp() * 1000

    # print "\n\nCollection Timestamp: %s" % collect_timestamp
    enforcer_result = self.get_legacyAdmin_enforcer()
    # print "Result from Rest API %s: %s" % (self._domain_enforcer_url, enforcer_result)
    if 'nuoca_collection_error' in enforcer_result:
      base_collection = {"TimeStamp": collect_timestamp,
                         'nuoca_collection_error': enforcer_result[
                           'nuoca_collection_error']}
      self._monitor_collect_queue.append(base_collection)
    regions_result = self.get_legacyAdmin_regions()
    # print "Result from Rest API %s: %s" % (self._regions_url, regions_result)
    if 'nuoca_collection_error' in regions_result:
      base_collection = {"TimeStamp": collect_timestamp,
                         'nuoca_collection_error': regions_result[
                           'nuoca_collection_error']}
      self._monitor_collect_queue.append(base_collection)
    else:
      region_count = 0
      for region in regions_result:
        base_collection = {"TimeStamp": collect_timestamp,
                           u'domainEnforcerEnabled': enforcer_result[
                             u'domainEnforcerEnabled']}
        for region_field in region_result_desired_fields:
          base_collection[region_field] = \
            regions_result[region_count][region_field]
        for host in region['hosts']:
          for host_field in host_result_desired_fields:
            base_collection["admin.%s" % host_field] = host[host_field]
          for host_tag_field in host_tags_desired_fields:
            base_collection["tag.%s" % host_tag_field] = \
              host['tags'][host_tag_field]
          for process in host['processes']:
            process_collection = {}
            for process_field in processes_desired_fields:
              process_collection["process.%s" % process_field] \
                = process[process_field]
            process_collection.update(base_collection)
            self._monitor_collect_queue.append(process_collection)
        for database in region['databases']:
          database_collection = {}
          for database_field in databases_desired_fields:
            if database_field in database_boolean_fields:
              database_collection["database.%s" % database_field] = \
                str(database[database_field]).lower()
            else:
              database_collection["database.%s" % database_field] = \
                str(database[database_field])
          database_collection.update(base_collection)
          self._monitor_collect_queue.append(database_collection)
        region_count += 1

  def _newAdmin_collector_thread(self, collect_timestamp):
    # Skipped collection of storage groups, kvstore.
    region_desired_fields = [u'id', u'name']
    archive_desired_fields = \
      [u'host', u'state', u'id', u'path', u'dbName']
    database_desired_fields = \
      [u'state', u'incarnation', u'name',
       u'hostAssignments', u'name', u'servers',
       u'databaseOptions', u'defaultRegionId']
    peers_desired_fields = \
      [u'connectedState', u'peerMemberState', u'isLocal',
       u'address', u'peerState', u'id']
    process_desired_fields = \
      [u'labels', u'pid', u'lastHeardFrom', u'port', u'regionName',
       u'hostname', u'nodeId', u'isExternalStartup', u'state', u'version',
       u'archiveId', u'type', u'regionId', u'host', u'address', u'startId',
       u'archiveDir', u'dbName', u'durableState', u'ipAddress']
    collect_timestamp = nuoca_gettimestamp() * 1000

    # Peers
    local_nuoadmin_id = None
    peers_result = self.get_newAdmin_peers()
    if not peers_result or 'data' not in peers_result:
      base_collection = {"TimeStamp": collect_timestamp,
                         'nuoca_collection_error':
                           "Unable to get peers information"}
      self._monitor_collect_queue.append(base_collection)
      return
    if 'nuoca_collection_error' in peers_result:
      base_collection = {"TimeStamp": collect_timestamp,
                         'nuoca_collection_error': peers_result[
                           'nuoca_collection_error']}
      self._monitor_collect_queue.append(base_collection)
      return
    # print "Result from Rest API %s: %s" % (self._peers_url, peers_result)
    for peer in peers_result['data']:
      if bool(peer['isLocal']):
        local_nuoadmin_id = peer['id']
    for peer in peers_result['data']:
      base_collection = {"TimeStamp": collect_timestamp,
                         'NuoCA.docType': 'peer',
                         'NuoCA.ApiServerID': local_nuoadmin_id}
      peer_collection = {}
      for peer_field in peers_desired_fields:
        collection_fieldname = '%s.%s' % (base_collection['NuoCA.docType'],
                                          peer_field)
        base_collection[collection_fieldname] = str(peer[peer_field])
      peer_collection.update(base_collection)
      self._monitor_collect_queue.append(peer_collection)

    # Regions
    regions_result = self.get_newAdmin_regions()
    # print "Result from Rest API %s: %s" % (self._regions_url, regions_result)
    for region in regions_result['data']:
      base_collection = {"TimeStamp": collect_timestamp,
                         'NuoCA.docType': 'region',
                         'NuoCA.ApiServerID': local_nuoadmin_id}
      region_collection = {}
      for region_field in region_desired_fields:
        collection_fieldname = '%s.%s' % (base_collection['NuoCA.docType'],
                                          region_field)
        base_collection[collection_fieldname] = str(region[region_field])
      region_collection.update(base_collection)
      self._monitor_collect_queue.append(region_collection)

    # Archives
    archives_result = self.get_newAdmin_archives()
    # print "Result from Rest API %s: %s" % (self._archives_url, archives_result)
    for archive in archives_result['data']:
      base_collection = {"TimeStamp": collect_timestamp,
                         'NuoCA.docType': 'archive',
                         'NuoCA.ApiServerID': local_nuoadmin_id}
      archive_collection = {}
      for archive_field in archive_desired_fields:
        collection_fieldname = '%s.%s' % (base_collection['NuoCA.docType'],
                                          archive_field)
        base_collection[collection_fieldname] = str(archive[archive_field])
      archive_collection.update(base_collection)
      self._monitor_collect_queue.append(archive_collection)

    # Databases
    databases_result = self.get_newAdmin_databases()
    # print "Result from Rest API %s: %s" % (self._databases_url, databases_result)
    if 'data' in databases_result:
      for db_item in databases_result['data']:
        base_collection = {"TimeStamp": collect_timestamp,
                           'NuoCA.docType': 'database',
                           'NuoCA.ApiServerID': local_nuoadmin_id}
        dbname = db_item['name']
        database_result = self.get_newAdmin_database(dbname)
        # print "Result from Rest API %s/%s: %s" % (self._databases_url,
        #                                          dbname, database_result)
        for database in database_result['data']:
          database_collection = {}
          for database_field in database_desired_fields:
            collection_fieldname = '%s.%s' % (base_collection['NuoCA.docType'],
                                              database_field)
            database_collection[collection_fieldname] = \
              str(database[database_field])
          # Deal with 'regions' like this
          collection_fieldname = '%s.%s' % (base_collection['NuoCA.docType'],
                                            'regions')
          database_collection[collection_fieldname] = \
            str(database['regions'].keys())
          database_collection.update(base_collection)
          self._monitor_collect_queue.append(database_collection)

    # Processes
    processes_result = self.get_newAdmin_processes()
    # print "Result from Rest API %s: %s" % (self._processes_url, processes_result)
    if 'data' in processes_result:
      for process in processes_result['data']:
        base_collection = {"TimeStamp": collect_timestamp,
                           'NuoCA.docType': 'process',
                           'NuoCA.ApiServerID': local_nuoadmin_id}
        process_results = {}
        for process_field in process_desired_fields:
          if process_field in process:
            collection_fieldname = '%s.%s' % (base_collection['NuoCA.docType'],
                                              process_field)
            process_results[collection_fieldname] = str(process[process_field])
        # Deal with 'options' like this
        for options_field in process['options']:
          collection_fieldname = '%s.%s.%s' % \
                                 (base_collection['NuoCA.docType'],
                                  'options', options_field)
          process_results[collection_fieldname] = \
            str(process['options'][options_field])
        process_results.update(base_collection)
        self._monitor_collect_queue.append(process_results)

  def _legacyAdmin_timer_thread(self):
    while self._enabled:
      collect_timestamp = self._admin_collect_sync.wait_for_next_interval()
      collect_thread = threading.Thread(
        target=self._legacyAdmin_collector_thread,
        args=(collect_timestamp,))
      collect_thread.daemon = True
      collect_thread.start()

  def _newAdmin_timer_thread(self):
    while self._enabled:
      collect_timestamp = self._admin_collect_sync.wait_for_next_interval()
      collect_thread = threading.Thread(target=self._newAdmin_collector_thread,
                                        args=(collect_timestamp,))
      collect_thread.daemon = True
      collect_thread.start()

  def startup(self, config=None):
    try:
      self._config = config
      if not config:
        nuoca_log(logging.ERROR, "NuoAdminMon plugin: missing config file.")
        return False

      # Validate the configuration.
      #required_config_items = ['admin_host']
      #if not self.has_required_config_items(config, required_config_items):
      #  return False

      # Don't reveal the domain password in the NuoCA log file.
      display_config = {}
      display_config.update(config)
      display_config['domain_password'] = ''
      nuoca_log(logging.INFO, "NuoAdminMonitor plugin config: %s" %
                str(display_config))

      if 'nuocaCollectionName' in config:
        self._nuocaCollectionName = config['nuocaCollectionName']

      # Legacy Admin
      if 'admin_host' in config:
        self._legacyAdmin_host = os.path.expandvars(config['admin_host'])
        if self._legacyAdmin_host:
          self._is_legacyAdmin = True
      if 'domain_username' in config:
        self._legacyAdmin_domain_username = \
          os.path.expandvars(config['domain_username'])
      if 'domain_password' in config:
        self._legacyAdmin_domain_password = \
          os.path.expandvars(config['domain_password'])
      if 'admin_rest_api_port' in config:
        if isinstance(config['admin_rest_api_port'], int):
          self._legacyAdmin_rest_api_port = config['admin_rest_api_port']
        else:
          self._legacyAdmin_rest_api_port = int(os.path.expandvars(
              config['admin_rest_api_port']))

      # New Admin
      if 'api_server' in config:
        self._newAdmin_api_server = os.path.expandvars(config['api_server'])
        self._is_legacyAdmin = False
      if 'client_key' in config:
        self._newAdmin_client_key = os.path.expandvars(config['client_key'])
      if 'server_cert' in config:
        self._newAdmin_server_cert = os.path.expandvars(config['server_cert'])

      if not self._legacyAdmin_host and not self._newAdmin_api_server:
        err_msg = "Config must have 'admin_host' for legacy Admin, " \
                  "or 'api_server' for new Admin"
        nuoca_log(logging.ERROR, "NuoAdminMonitor Plugin: %s" % err_msg)
        return False

      if 'admin_collect_interval' in config:
        if isinstance(config['admin_collect_interval'], int):
          self._admin_collect_interval = config['admin_collect_interval']
        else:
          self._admin_collect_interval = \
            os.path.expandvars(config['admin_collect_interval'])
      if 'admin_collect_timeout' in config:
        if isinstance(config['admin_collect_timeout'], int):
          self._admin_collect_timeout = config['admin_collect_timeout']
        else:
          self._admin_collect_timeout = \
            os.path.expandvars(config['admin_collect_timeout'])
      if 'database_regex_pattern' in config:
        self._database_regex_pattern = config['database_regex_pattern']
      if 'host_uuid_shortname' in config:
        self._host_uuid_shortname = config['host_uuid_shortname']

      if self._is_legacyAdmin:
        self._legacyAdmin_base_url = "http://%s:%d/api/2" % \
                                     (self._legacyAdmin_host,
                                      self._legacyAdmin_rest_api_port)
        self._legacyAdmin_domain_enforcer_url = "%s/domain/enforcer" % \
                                                self._legacyAdmin_base_url
        self._newAdmin_regions_url = "%s/regions" % self._legacyAdmin_base_url
        self._legacyAdmin_auth = \
          HTTPBasicAuth(self._legacyAdmin_domain_username,
                        self._legacyAdmin_domain_password)
        nuoca_start_ts = None
        if 'nuoca_start_ts' in config:
          nuoca_start_ts = config['nuoca_start_ts']
        self._admin_collect_sync = IntervalSync(self._admin_collect_interval,
                                                seed_ts=nuoca_start_ts)

        self._enabled = True
        self._timer_thrd = \
          threading.Thread(target=self._legacyAdmin_timer_thread)
        self._timer_thrd.daemon = True
        self._timer_thrd.start()
      else:
        admin_url = self._get_newAdmin_url()
        self._newAdmin_base_url = "%s/api/1" % admin_url
        self._newAdmin_regions_url = "%s/regions" % self._newAdmin_base_url
        self._newAdmin_archives_url = "%s/archives" % self._newAdmin_base_url
        self._newAdmin_databases_url = "%s/databases" % self._newAdmin_base_url
        self._newAdmin_kvstore_url = "%s/kvstore" % self._newAdmin_base_url
        self._newAdmin_peers_url = "%s/peers" % self._newAdmin_base_url
        self._newAdmin_processes_url = "%s/processes" % self._newAdmin_base_url

        nuoca_start_ts = None
        if 'nuoca_start_ts' in config:
          nuoca_start_ts = config['nuoca_start_ts']
        self._admin_collect_sync = IntervalSync(self._admin_collect_interval,
                                                seed_ts=nuoca_start_ts)

        self._enabled = True
        self._timer_thrd = threading.Thread(target=self._newAdmin_timer_thread)
        self._timer_thrd.daemon = True
        self._timer_thrd.start()

      return True
    except Exception as e:
      nuoca_log(logging.ERROR, "NuoAdminMonitor Plugin: %s" % str(e))
      return False

  def shutdown(self):
    self.enabled = False
    pass

  def collect(self, collection_interval):
    rval = None
    try:
      nuoca_log(logging.DEBUG, "Called collect() in NuoAdminMonitor Plugin process")
      base_values = super(NuoAdminMonitorPlugin, self).collect(collection_interval)
      collection_count = len(self.monitor_collect_queue)
      if not collection_count:
        return rval

      rval = []
      for i in range(collection_count):
        collected_dict = self.monitor_collect_queue.pop(0)
        if self._nuocaCollectionName:
          collected_dict['nuocaCollectionName'] = self._nuocaCollectionName
        if 'process.dbname' in collected_dict:
          m = re.search(self._database_regex_pattern, collected_dict['process.dbname'])
          if m:
            rval.append(collected_dict)
        else:
          rval.append(collected_dict)
    except Exception as e:
      nuoca_log(logging.ERROR, str(e))
    return rval



# Copyright (c) 2017-2020, NuoDB, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of NuoDB, Inc. nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUODB, INC. BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
from kafka import KafkaProducer
import json

from pynuoca.nuoca_plugin import NuocaMPOutputPlugin
from pynuoca.nuoca_util import nuoca_log


class KafkaProducerPlugin(NuocaMPOutputPlugin):
    def __init__(self, parent_pipe, config=None):
        super(KafkaProducerPlugin, self).__init__(parent_pipe, 'KafkaProducer')
        self._config = config
        self._producer = None
        self._defaulttopic = None

    def startup(self, config=None):
        try:
            self._config = config

            servers = config.get('servers', 'localhost:9092')
            self._defaulttopic = config.get('defaulttopic', None)
            print servers
            self._producer = KafkaProducer(bootstrap_servers=servers,
                                           value_serializer=lambda v: json.dumps(v).encode('utf-8')
                                           )
            return True
        except Exception as e:
            nuoca_log(logging.ERROR, str(e))

    def shutdown(self):
        pass

    def store(self, ts_values):
        rval = None
        try:
            nuoca_log(logging.DEBUG,
                      "Called store() in KafkaProducerPlugin process")
            rval = super(KafkaProducerPlugin, self).store(ts_values)
            topic = self._defaulttopic
            if topic is not None:
                if type(ts_values) is list:
                    for svalues in ts_values:
                        self._producer.send(topic, svalues)
                else:
                    self._producer.send(topic, ts_values)
        except Exception as e:
            nuoca_log(logging.ERROR, str(e))
        return rval



# (C) Copyright NuoDB, Inc. 2017-2020
#
# This source code is licensed under the MIT license found in the LICENSE
# file in the root directory of this source tree.

import os
import logging
import requests
import json

from pynuoca.nuoca_plugin import NuocaMPOutputPlugin
from pynuoca.nuoca_util import nuoca_log


class RestClientPlugin(NuocaMPOutputPlugin):
    def __init__(self, parent_pipe, config=None):
        super(RestClientPlugin, self).__init__(parent_pipe, 'RestClient')
        self._config = config
        self._url = None
        self._auth_token = None

    def startup(self, config=None):
        try:
            self._config = config
            required_config_items = ['url']
            if not self.has_required_config_items(config, required_config_items):
                return False
            self._url = os.path.expandvars(config['url'])
            if 'auth_token' in config:
                self._auth_token = os.path.expandvars(config['auth_token'])
            return True
        except Exception as e:
            nuoca_log(logging.ERROR, str(e))

    def shutdown(self):
        pass

    def store(self, ts_values):
        rval = None
        try:
            nuoca_log(logging.DEBUG,
                      "Called store() in RestClientPlugin process")
            rval = super(RestClientPlugin, self).store(ts_values)
            headers = {"content-type": "application/json"}
            if self._auth_token:
                headers['Authorization'] = "Bearer %s" % self._auth_token
            requests.post(self._url, json=ts_values,
                          headers=headers)
        except Exception as e:
            nuoca_log(logging.ERROR, str(e))
        return rval



# (C) Copyright NuoDB, Inc. 2017-2020
#
# This source code is licensed under the MIT license found in the LICENSE
# file in the root directory of this source tree.

import os
import traceback
import logging
from pynuoca.nuoca_util import nuoca_gettimestamp, nuoca_log
from yapsy.IMultiprocessChildPlugin import IMultiprocessChildPlugin


class NuocaMPPlugin(IMultiprocessChildPlugin):
    """
    Base class for NuoCA Multi-Process Plugins
    NuoCA plugins are based on Python Yapsy: http://yapsy.sourceforge.net/
    """

    def __init__(self, parent_pipe, plugin_name, plugin_type):
        """
        :param parent_pipe: Pipe setup by Yaspy
        :param plugin_name: Name of the plugin.
        :type plugin_name: ``str``
        :param plugin_type: Plugin type: one of: "Input", "Output", "Transform"
        :type plugin_type: ``str``
        """
        nuoca_log(logging.INFO, "Creating plugin: %s" % plugin_name)
        self._parent_pipe = parent_pipe
        self._plugin_name = plugin_name
        self._type = plugin_type
        self._enabled = False
        IMultiprocessChildPlugin.__init__(self, parent_pipe=parent_pipe)

    @property
    def plugin_name(self):
        return self._plugin_name

    @property
    def type(self):
        return self._type

    @property
    def enabled(self):
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        self._enabled = value

    def activate(self):
        nuoca_log(logging.INFO, "Activate plugin: %s" % self.name)
        super(NuocaMPPlugin, self).activate()
        self._enabled = True

    def startup(self, config):
        raise NotImplementedError("Child class must implement startup method")

    def shutdown(self):
        raise NotImplementedError("Child class must implement shutdown method")

    def deactivate(self):
        nuoca_log(logging.INFO, "Deactivate plugin: %s" % self.name)
        super(NuocaMPPlugin, self).deactivate()
        self._enabled = False

    def has_required_config_items(self, config, required_config_items):
        """
        Check that the configuration contains the required items.

        :param config: Configuration to check
        :type config: ``dict``

        :param required_config_items: List of required configuration item names
        :type required_config_items: ``List[str]``

        :return: False if config or required items are missing.  Otherwise
        return True.
        """
        if not config:
            nuoca_log(logging.ERROR, "%s plugin: missing config file." %
                      self._plugin_name)
            return False
        for config_item in required_config_items:
            if config_item not in config:
                nuoca_log(logging.ERROR, "%s plugin: '%s' missing from config." %
                          (self._plugin_name, config_item))
                return False
        return True


class NuocaMPInputPlugin(NuocaMPPlugin):
    """
    NuoCA Multi-Process Input Plugin

    This is a base class for ALL NuoCA Input Plugins.

    All NuoCA Input Plugins must do:
      1) Implement a Class that derives from NuocaMPInputPlugin
      2) call this __init__ function from the Plugin's __init__.
      3) Implement a collect() method that calls this collect() method.
    """

    def __init__(self, parent_pipe, plugin_name):
        """
        :param parent_pipe: Provided by Yapsy
        :param plugin_name: Plugin name
        :type plugin_name: ``str``
        """
        super(NuocaMPInputPlugin, self).__init__(parent_pipe, plugin_name, "Input")
        self._collection_name = plugin_name

    def _send_response(self, status_code, err_msg=None, resp_dict=None):
        response = {'status_code': status_code}
        if err_msg:
            response['error_msg'] = err_msg
        if resp_dict:
            response['resp_values'] = resp_dict
        self.parent_pipe.send(response)

    def run(self):
        """
        This function is called by Yapsy
        """
        self.enabled = True
        nuoca_log(logging.INFO, "Plugin: %s,  PID: %d" % (self.plugin_name, os.getpid()))
        while self.enabled:
            try:
                request_from_parent = self.parent_pipe.recv()
                if not request_from_parent:
                    self._send_response(2, "Empty request from parent in Plugin: %s"
                                        % self.name)
                    continue
                if 'action' not in request_from_parent:
                    self._send_response(2, "Action missing from request in Plugin: %s"
                                        % self.name)
                    continue
                action = request_from_parent['action']
                if action == 'collect':
                    collection_interval = request_from_parent['collection_interval']
                    collected_values = self.collect(collection_interval)
                    self._send_response(0, None, {'collected_values': collected_values})
                    continue
                elif action == 'startup':
                    config = request_from_parent['config']
                    if config:
                        if 'nuocaCollectionName' in config:
                            self._collection_name = config['nuocaCollectionName']
                        if 'environmentVariables' in config:
                            for env_var in config['environmentVariables']:
                                os.environ[env_var['name']] = env_var['value']
                                msg = "Env: %s=%s" % (env_var['name'], env_var['value'])
                                nuoca_log(logging.INFO, msg)

                    startup_rval = self.startup(config)
                    self._send_response(int(not startup_rval))
                    continue
                elif action == 'shutdown':
                    self.shutdown()
                    continue
                elif action == 'exit':
                    self.enabled = False
                    self._send_response(0, None, {'goodbye': 'world'})
                    continue
                else:
                    self._send_response(2, "Action %s unknown in Plugin: %s"
                                        % (action, self.name))
                    continue

            except Exception as e:
                err_msg = "Unhandled exception: %s\n%s" % (e, traceback.format_exc())
                self._send_response(1, err_msg)

    def deactivate(self):
        super(NuocaMPInputPlugin, self).deactivate()

    def collect(self, collection_interval):
        """
        NuoCA Input Plugins must implement their own collect() method.  The
        NuoCA Input Plugins must call this collect() method at the top of
        their collect() method.  This method returns a python dictionary
        of values that each NuoCA Input Plugin should add to their own
        time-series information.

        Note: The plugin's collect() method must return a Python list
        of one or more Python dictionaries that contain time-series values.

        NuoCA will call the collect() method at the end of each Collection
        Interval.

        :param collection_interval: Collection Interval
        :type collection_interval: ``int``
        :return: time-series values
        :type: ``dict``
        """
        rval = {'collect_timestamp': nuoca_gettimestamp() * 1000}
        return rval


class NuocaMPOutputPlugin(NuocaMPPlugin):
    """
    NuoCA Multi-Process Output Plugin

    This is a base class for ALL NuoCA Output Plugins.

    All NuoCA Output Plugins must do:
      1) Implement a Class that derives from NuocaMPOutputPlugin
      2) call this __init__ function from the Plugin's __init__.
      3) Implement a store() method that calls this store() method.
    """

    def __init__(self, parent_pipe, plugin_name):
        super(NuocaMPOutputPlugin, self).__init__(parent_pipe, plugin_name,
                                                  "Output")

    def _send_response(self, status_code, err_msg=None, resp_dict=None):
        response = {'status_code': status_code}
        if err_msg:
            response['error_msg'] = err_msg
        if resp_dict:
            response['resp_values'] = resp_dict
        self.parent_pipe.send(response)

    def run(self):
        """
        This function is called by Yapsy
        """
        self.enabled = True
        nuoca_log(logging.INFO, "Plugin: %s,  PID: %d" % (self.plugin_name, os.getpid()))
        while self.enabled:
            try:
                request_from_parent = self.parent_pipe.recv()
                if not request_from_parent:
                    self._send_response(2, "Empty request from parent in Plugin: %s"
                                        % self.name)
                    continue
                if 'action' not in request_from_parent:
                    self._send_response(2, "Action missing from request in Plugin: %s"
                                        % self.name)
                    continue
                action = request_from_parent['action']
                if action == 'store':
                    ts_values = request_from_parent['ts_values']
                    resp_from_store = self.store(ts_values)
                    self._send_response(0, None, resp_from_store)
                    continue
                elif action == 'startup':
                    config = request_from_parent['config']
                    if config:
                        if 'environmentVariables' in config:
                            for env_var in config['environmentVariables']:
                                os.environ[env_var['name']] = env_var['value']
                                msg = "Env: %s=%s" % (env_var['name'], env_var['value'])
                                nuoca_log(logging.INFO, msg)

                    startup_rval = self.startup(config)
                    self._send_response(int(not startup_rval))
                    continue
                elif action == 'shutdown':
                    self.shutdown()
                    continue
                elif action == 'exit':
                    self.enabled = False
                    self._send_response(0, None, {'goodbye': 'world'})
                    continue
                else:
                    self._send_response(2, "Action %s unknown in Plugin: %s"
                                        % (action, self.name))
                    continue
            except Exception as e:
                err_msg = "Unhandled exception: %s\n%s" % (e, traceback.format_exc())
                self._send_response(1, err_msg)

    def store(self, ts_values):
        pass


class NuocaMPTransformPlugin(NuocaMPPlugin):
    """
    NuoCA Multi-Process Transformation Plugin
    """

    def __init__(self, parent_pipe, plugin_name):
        super(NuocaMPTransformPlugin, self).__init__(parent_pipe,
                                                     plugin_name, "Transform")
        raise Exception("Not Yet Implemented")



# (C) Copyright NuoDB, Inc. 2017-2020
#
# This source code is licensed under the MIT license found in the LICENSE
# file in the root directory of this source tree.

import os
import click
import socket
import signal
import traceback
import yaml
from pynuoca.nuoca_util import *
from yapsy.MultiprocessPluginManager import MultiprocessPluginManager
from pynuoca.nuoca_plugin import NuocaMPInputPlugin, NuocaMPOutputPlugin, \
    NuocaMPTransformPlugin
from nuoca_config import NuocaConfig

nuoca_obj = None


class NuoCA(object):
    """
    NuoDB Collection Agent
    """

    def __init__(self, config_file=None, collection_interval=30,
                 plugin_dir=None, starttime=None, mode=None,
                 verbose=False, self_test=False, log_level=logging.INFO,
                 output_values=None):
        """
        :param config_file: Path to NuoCA configuration file.
        :type config_file: ``str``

        :param collection_interval: Collection Interval in seconds
        :type collection_interval: ``int``

        :param plugin_dir: Path to NuoCA Plugin Directory
        :type plugin_dir: ``str``

        :param starttime: Epoch timestamp of start time of the first collection.
        :type starttime: ``int``, ``None``

        :param mode: Run mode; [coach | insights]
        :type plugin_dir: ``str``

        :param verbose: Flag to indicate printing of verbose messages to stdout.
        :type verbose: ``bool``

        :param self_test: Flag to indicate a 5 loop self test.
        :type self_test: ``bool``

        :param log_level: Python logging level
        :type log_level: ``logging.level``

        :param output_values: list of strings parsable by utils.parse_keyval_list()
        :type output_values: `list` of `str`
        """

        self._starttime = starttime
        if self._starttime:
            # Make sure that starttime is now or in the future.
            current_timestamp = nuoca_gettimestamp()
            if current_timestamp >= self._starttime:
                msg = "starttime must be now or in the future."
                nuoca_log(logging.ERROR, msg)
                raise AttributeError(msg)

        self._nuoca_cfgdir = None
        self.get_nuoca_cfgdir()
        self._nuoca_settings = self.read_nuoca_settings()

        # override default log level
        log_level = self._get_settings_loglevel()

        # collect nuoca.log?
        self._collect_nuoca_log = False
        collect_nuoca_log = self.get_nuoca_settings_value('collect.nuoca.log')
        if collect_nuoca_log:
            self._collect_nuoca_log = True

        # logCollectionCount?
        self._log_collection_counts = False
        log_collection_counts = self.get_nuoca_settings_value('logCollectionCounts')
        if log_collection_counts:
            self._log_collection_counts = True

        # nuoca config file
        self._config = NuocaConfig(config_file)
        self._hostname = socket.gethostname()
        logfile = os.path.expandvars(self._config.NUOCA_LOGFILE)
        initialize_logger(logfile, self._collect_nuoca_log)

        nuoca_set_log_level(log_level)
        nuoca_log(logging.INFO, "nuoca server init.")
        nuoca_log(logging.INFO, "NuoCA, PID: %d" % os.getpid())
        self._collection_interval = collection_interval

        # Override the command line/default NuoCA collection interval if
        # NUOCA_COLLECTION_INTERVAL is specified in the NuoCA config file.
        if self._config.NUOCA_COLLECTION_INTERVAL:
            self._collection_interval = \
                int(self._config.NUOCA_COLLECTION_INTERVAL)

        if not starttime:
            self._starttime = None
        else:
            self._starttime = int(starttime)
        self._plugin_topdir = plugin_dir
        self._enabled = True
        self._mode = mode
        self._verbose = verbose  # Used to make stdout verbose.
        self._self_test = self_test
        self._output_values = parse_keyval_list(output_values)
        if self._output_values:
            for output_val in self._output_values:
                nuoca_log(logging.INFO, "Output Key: '%s' set to Value: '%s'" %
                          (output_val, self._output_values[output_val]))

        # The following self._*_plugins are dictionaries of two element
        # tuples in the form: (plugin object, plugin configuration) keyed
        # by the plugin name.
        self._input_plugins = {}
        self._output_plugins = {}
        self._transform_plugins = {}

        if not self._plugin_topdir:
            self._plugin_topdir = os.path.join(get_nuoca_sourcedir(), "plugins")
        nuoca_log(logging.INFO, "plugin dir: %s" % self._plugin_topdir)

        input_plugin_dir = os.path.join(self._plugin_topdir, "input")
        output_plugin_dir = os.path.join(self._plugin_topdir, "output")
        transform_plugin_dir = os.path.join(self._plugin_topdir, "transform")
        self._plugin_directories = [input_plugin_dir,
                                    output_plugin_dir,
                                    transform_plugin_dir]

    def get_nuoca_cfgdir(self):
        if self._nuoca_cfgdir:
            return self._nuoca_cfgdir
        try:
            self._nuoca_cfgdir = os.path.join(os.path.expandvars('$NUOCA_HOME'), 'etc')
        except:
            pass
        return self._nuoca_cfgdir

    def read_nuoca_settings(self):
        settings = None
        try:
            if self._nuoca_cfgdir:
                settings_file = os.path.join(self._nuoca_cfgdir, 'nuoca_settings.yml')
                nuoca_log(logging.INFO, "Using CONFIG file %s" % settings_file)
                with open(settings_file, 'r') as f:
                    settings = yaml.safe_load(f.read())
                    print settings
                f.close()
        except:
            pass
        return settings

    def get_nuoca_settings_value(self, variable):
        if not self._nuoca_settings:
            return None
        try:
            value = self._nuoca_settings[variable]
            return value
        except:
            return None

    @property
    def config(self):
        return self._config

    def _get_settings_loglevel(self):
        loglevel = None
        try:
            loglevel_str = self.get_nuoca_settings_value('nuoca.logging.level')
            level_mappings = {
                'DEBUG': logging.DEBUG,
                'INFO': logging.INFO,
                'WARNING': logging.WARNING,
                'ERROR': logging.ERROR,
                'CRITICAL': logging.CRITICAL
            }
            loglevel = level_mappings[loglevel_str]
        except:
            pass
        if not loglevel:
            loglevel = logging.INFO
        return loglevel

    def _collection_cycle(self, collection_time):
        """
        _collection_cycle is called at the end of each Collection
        Interval.
        """
        nuoca_log(logging.INFO, "Starting collection interval: %s" %
                  collection_time)
        collected_inputs = self._collect_inputs()
        for list_item in collected_inputs:
            list_item['collection_interval'] = self._collection_interval
            if 'timestamp' not in list_item:
                list_item['timestamp'] = collection_time
            list_item['NuoCA.Hostname'] = self._hostname
        # TODO Transformations
        self._store_outputs(collected_inputs)

    def _get_activated_input_plugins(self):
        """
        Get a list of "activated" input plugins
        """
        input_list = self.manager.getPluginsOfCategory('Input')
        activated_list = [x for x in input_list if x.is_activated]
        return activated_list

    def _get_deactivated_input_plugins(self):
        """
        Get a list of "deactivated" input plugins
        """
        input_list = self.manager.getPluginsOfCategory('Input')
        deactivated_list = [x for x in input_list if not x.is_activated]
        return deactivated_list

    def _get_activated_output_plugins(self):
        """
        Get a list of "activated" output plugins
        """
        output_list = self.manager.getPluginsOfCategory('Output')
        activated_list = [x for x in output_list if x.is_activated]
        return activated_list

    def _get_deactivated_output_plugins(self):
        """
        Get a list of "activated" output plugins
        """
        output_list = self.manager.getPluginsOfCategory('Output')
        deactivated_list = [x for x in output_list if not x.is_activated]
        return deactivated_list

    def _get_plugin_respose(self, a_plugin):
        """
        Get the response message from the plugin
        :return: Response dictionary if successful, otherwise None.
        """
        plugin_obj = a_plugin.plugin_object
        # noinspection PyBroadException
        try:
            if plugin_obj.child_pipe.poll(self.config.PLUGIN_PIPE_TIMEOUT):
                response = plugin_obj.child_pipe.recv()
                if self._verbose:
                    print("%s:%s" % (a_plugin.name, response))
            else:
                nuoca_log(logging.ERROR,
                          "NuoCA._get_plugin_respose: "
                          "Timeout collecting response values from plugin: %s"
                          % a_plugin.name)
                return None

        except Exception as e:
            nuoca_log(logging.ERROR,
                      "NuoCA._get_plugin_respose: "
                      "Unable to collect response from plugin: %s\n%s"
                      % (a_plugin.name, str(e)))
            return None

        # noinspection PyBroadException
        try:
            if not response:
                nuoca_log(logging.ERROR,
                          "NuoCA._get_plugin_respose: "
                          "Missing response from plugin: %s"
                          % a_plugin.name)
                return None
            if 'status_code' not in response:
                nuoca_log(logging.ERROR,
                          "NuoCA._get_plugin_respose: "
                          "status_code missing from plugin response: %s"
                          % a_plugin.name)
                return None

        except Exception as e:
            nuoca_log(logging.ERROR,
                      "NuoCA._get_plugin_respose: Error attempting to collect"
                      " response from plugin: %s\n%s"
                      % (a_plugin.name, str(e)))
            return None

        return response

    def _startup_plugin(self, a_plugin, config=None):
        """
        Send start message to plugin.
        :param a_plugin: The plugin
        :param config: NuoCA Configuration
        :type config: ``dict``
        """
        response = None
        nuoca_log(logging.INFO, "Called to start plugin: %s" % a_plugin.name)
        plugin_msg = {'action': 'startup', 'config': config}
        try:
            a_plugin.plugin_object.child_pipe.send(plugin_msg)
        except Exception as e:
            nuoca_log(logging.ERROR,
                      "Unable to send %s message to plugin: %s\n%s"
                      % (plugin_msg, a_plugin.name, str(e)))

        try:
            response = self._get_plugin_respose(a_plugin)
        except Exception as e:
            nuoca_log(logging.ERROR,
                      "Problem with response on %s message to plugin: %s\n%s"
                      % (plugin_msg, a_plugin.name, str(e)))
        if not response or response['status_code'] != 0:
            nuoca_log(logging.ERROR,
                      "Disabling plugin that failed to startup: %s"
                      % a_plugin.name)
            self.manager.deactivatePluginByName(a_plugin.name, a_plugin.category)
            self._shutdown_plugin(a_plugin)

    @staticmethod
    def _exit_plugin(a_plugin):
        """
        Send Exit message to plugin.
        :param a_plugin: The plugin
        """
        nuoca_log(logging.INFO, "Called to exit plugin: %s" % a_plugin.name)
        plugin_msg = {'action': 'exit'}
        try:
            a_plugin.plugin_object.child_pipe.send(plugin_msg)
        except Exception as e:
            nuoca_log(logging.ERROR,
                      "Unable to send %s message to plugin: %s\n%s"
                      % (plugin_msg, a_plugin.name, str(e)))

    @staticmethod
    def _shutdown_plugin(a_plugin):
        """
        Send stop message to plugin.

        :param a_plugin: The plugin
        :type a_plugin: NuocaMPPlugin
        """
        nuoca_log(logging.INFO, "Called to shutdown plugin: %s" % a_plugin.name)
        plugin_msg = {'action': 'shutdown'}
        try:
            a_plugin.plugin_object.child_pipe.send(plugin_msg)
        except Exception as e:
            nuoca_log(logging.ERROR,
                      "Unable to send %s message to plugin: %s\n%s"
                      % (plugin_msg, a_plugin.name, str(e)))

    def _collect_inputs(self):
        """
        Collect time-series data from each activated plugin.
        :return: ``dict`` of time-series data
        """
        # TODO - Use Threads so that we can do concurrent collection.
        plugin_msg = {'action': 'collect',
                      'collection_interval': self._collection_interval}
        rval = []
        activated_plugins = self._get_activated_input_plugins()
        for a_plugin in activated_plugins:
            # noinspection PyBroadException
            try:
                a_plugin.plugin_object.child_pipe.send(plugin_msg)
            except Exception as e:
                nuoca_log(logging.ERROR,
                          "NuoCA._collect_inputs: "
                          "Unable to send %s message to plugin: %s\n%s"
                          % (plugin_msg, a_plugin.name, str(e)))

        for a_plugin in activated_plugins:
            response = self._get_plugin_respose(a_plugin)
            if not response:
                continue
            if 'resp_values' not in response:
                continue

            resp_values = response['resp_values']

            # noinspection PyBroadException
            try:
                if 'collected_values' not in resp_values:
                    nuoca_log(logging.ERROR,
                              "NuoCA._collect_inputs: "
                              "'Collected_Values' missing in response from plugin: %s"
                              % a_plugin.name)
                    continue
                if not resp_values['collected_values']:
                    nuoca_log(logging.DEBUG,
                              "No time-series values were collected from plugin: %s"
                              % a_plugin.name)
                    if self._log_collection_counts:
                        nuoca_log(logging.INFO,
                                  "%s Collection Count: 0" % a_plugin.name)
                    continue
                if type(resp_values['collected_values']) is not list:
                    nuoca_log(logging.ERROR,
                              "NuoCA._collect_inputs: "
                              "'Collected_Values' is not a list in "
                              "response from plugin: %s"
                              % a_plugin.name)
                    continue

                list_count = len(resp_values['collected_values'])
                if self._log_collection_counts:
                    nuoca_log(logging.INFO,
                              "%s Collection Count: %d" % (a_plugin.name, list_count))

                for list_index in range(list_count):
                    new_values = {}
                    key_prefix = a_plugin.name
                    collected_dict = resp_values['collected_values'][list_index]
                    if 'nuocaCollectionName' in collected_dict:
                        key_prefix = collected_dict['nuocaCollectionName']
                        del collected_dict['nuocaCollectionName']
                    for collected_item in collected_dict:
                        key_name = key_prefix + '.' + collected_item
                        new_values[key_name] = collected_dict[collected_item]
                        if collected_item == 'TimeStamp':
                            new_values['timestamp'] = int(collected_dict[collected_item])
                    new_values['NuoCA.plugin_name'] = key_prefix
                    for input_plugin_config in self._config.INPUT_PLUGINS:
                        if input_plugin_config.keys()[0] == a_plugin.name:
                            if 'NuoCA.route' in input_plugin_config.values()[0]:
                                new_values['NuoCA.route'] = input_plugin_config.values()[0]['NuoCA.route']
                    if self._output_values:
                        new_values.update(self._output_values)
                    rval.append(new_values)
            except Exception as e:
                nuoca_log(logging.ERROR,
                          "NuoCA._collect_inputs: "
                          "Error attempting to collect"
                          " response from plugin: %s\n%s"
                          % (a_plugin.name, str(e)))

        try:
            deactivated_plugins = self._get_deactivated_input_plugins()
            for a_plugin in deactivated_plugins:
                for a_config in self._config.INPUT_PLUGINS:
                    a_config_name = a_config.keys()[0]
                    if a_config_name == a_plugin.name:
                        self._activate_and_startup_input_plugin(a_config)
            deactivated_plugins = self._get_deactivated_output_plugins()
            for a_plugin in deactivated_plugins:
                for a_config in self._config.OUTPUT_PLUGINS:
                    a_config_name = a_config.keys()[0]
                    if a_config_name == a_plugin.name:
                        self._activate_and_startup_output_plugin(a_config)
        except Exception as e:
            nuoca_log(logging.WARNING,
                      "Unable to reactivate deactived plugins during collection: %s"
                      % str(e))

        if self._collect_nuoca_log:
            try:
                nuoca_collect_loghandler = get_nuoca_collect_loghandler()
                items = nuoca_collect_loghandler.collect(self._collection_interval)
                for item in items:
                    if self._output_values:
                        item.update(self._output_values)
                    item['NuoCA.plugin_name'] = 'NuoCALog'
                    rval.append(item)
            except:
                pass

        return rval

    def _store_outputs(self, collected_inputs):
        if not collected_inputs:
            return
        rval = {}
        activated_plugins = self._get_activated_output_plugins()
        output_plugins_used = []
        for a_plugin in activated_plugins:
            outputs = collected_inputs
            # noinspection PyBroadException
            try:
                if self._config.is_routing:
                    outputs = []
                    for input_item in collected_inputs:
                        if 'NuoCA.route' in input_item:
                            if a_plugin.name in input_item['NuoCA.route'].split(','):
                                outputs.append(input_item)
                        else:
                            outputs.append(input_item)
                if len(outputs) > 0:
                    output_plugins_used.append(a_plugin)
                    plugin_msg = {'action': 'store', 'ts_values': outputs}
                    a_plugin.plugin_object.child_pipe.send(plugin_msg)
            except Exception as e:
                nuoca_log(logging.ERROR,
                          "Unable to send 'Store' message to plugin: %s\n%s"
                          % (a_plugin.name, str(e)))

        for a_plugin in output_plugins_used:
            resp_values = self._get_plugin_respose(a_plugin)
            if not resp_values:
                continue

        return rval

    def _create_plugin_manager(self):
        self.manager = MultiprocessPluginManager(
            directories_list=self._plugin_directories,
            plugin_info_ext="multiprocess-plugin")
        self.manager.setCategoriesFilter({
            "Input": NuocaMPInputPlugin,
            "Output": NuocaMPOutputPlugin,
            "Transform": NuocaMPTransformPlugin
        })

    def _activate_and_startup_input_plugin(self, input_plugin):
        input_plugin_name = input_plugin.keys()[0]
        a_plugin = self.manager.getPluginByName(input_plugin_name, 'Input')
        if a_plugin and not a_plugin.is_activated:
            if not self.manager.activatePluginByName(input_plugin_name, 'Input'):
                err_msg = "Cannot activate input plugin: '%s', Skipping." % \
                          input_plugin_name
                nuoca_log(logging.WARNING, err_msg)
            else:
                input_plugin_config = input_plugin.values()[0]
                if not input_plugin_config:
                    input_plugin_config = {}
                input_plugin_config['nuoca_start_ts'] = self._starttime
                input_plugin_config['nuoca_collection_interval'] = \
                    self._collection_interval
                self._startup_plugin(a_plugin, input_plugin_config)
                self._input_plugins[input_plugin_name] = (a_plugin,
                                                          input_plugin_config)

    def _activate_and_startup_output_plugin(self, output_plugin):
        output_plugin_name = output_plugin.keys()[0]
        a_plugin = self.manager.getPluginByName(output_plugin_name, 'Output')
        if a_plugin and not a_plugin.is_activated:
            if not self.manager.activatePluginByName(output_plugin_name, 'Output'):
                err_msg = "Cannot activate output plugin: '%s', Skipping." % \
                          output_plugin_name
                nuoca_log(logging.WARNING, err_msg)
            else:
                output_plugin_config = output_plugin.values()[0]
                if not output_plugin_config:
                    output_plugin_config = {}
                output_plugin_config['nuoca_start_ts'] = self._starttime
                output_plugin_config['nuoca_collection_interval'] = \
                    self._collection_interval
                self._startup_plugin(a_plugin, output_plugin_config)
                self._output_plugins[output_plugin_name] = (a_plugin,
                                                            output_plugin_config)

    # Activate plugins and call the plugin's startup() method.
    def _activate_and_startup_plugins(self):
        for input_plugin in self.config.INPUT_PLUGINS:
            self._activate_and_startup_input_plugin(input_plugin)
        for output_plugin in self.config.OUTPUT_PLUGINS:
            self._activate_and_startup_output_plugin(output_plugin)
        # TODO Transform Plugins

    # test if the plugin name is configured in NuoCA.
    def _is_plugin_name_configured(self, name):
        for configured_input in self.config.INPUT_PLUGINS:
            if name in configured_input:
                return True
        for configured_input in self.config.OUTPUT_PLUGINS:
            if name in configured_input:
                return True
        for configured_input in self.config.TRANSFORM_PLUGINS:
            if name in configured_input:
                return True
        return False

    # activate only the NuoCA configured plugins.
    def _activate_configured_plugins(self):
        self.manager.locatePlugins()
        # get a list of ALL plugin candidates
        plugin_candidates = self.manager.getPluginCandidates()
        for candidate in plugin_candidates:
            plugin_configured = self._is_plugin_name_configured(candidate[2].name)
            if not plugin_configured:
                # Remove this plugin candidate because it is no configued by NuoCA
                self.manager.removePluginCandidate(candidate)
        self.manager.loadPlugins()
        self._activate_and_startup_plugins()

    def _shutdown_all_plugins(self):
        for input_plugin in self._input_plugins:
            self.manager.deactivatePluginByName(input_plugin, 'Input')
            a_plugin = self.manager.getPluginByName(input_plugin, 'Input')
            self._shutdown_plugin(a_plugin)
        for output_plugin in self._output_plugins:
            self.manager.deactivatePluginByName(output_plugin, 'Output')
            a_plugin = self.manager.getPluginByName(output_plugin, 'Output')
            self._shutdown_plugin(a_plugin)
        # TODO Transform Plugins

    @staticmethod
    def kill_all_plugin_processes(manager, timeout=5):
        """
        Kill any plugin processes that were left running after waiting up to
        the timeout value..

        :param manager: MultiprocessPluginManager
        :type manager: MultiprocessPluginManager

        :param timeout: Maximum time to wait (in seconds) for the process to
        self exit before killing.
        :type timeout: ``int``

        """
        if not manager:
            return
        all_plugins = manager.getAllPlugins()
        wait_count = timeout
        for a_plugin in all_plugins:
            while a_plugin.plugin_object.proc.is_alive() and wait_count > 0:
                time.sleep(1)
                wait_count -= 1
            if a_plugin.plugin_object.proc.is_alive():
                nuoca_log(logging.INFO, "Killing plugin subprocess: %s" % a_plugin)
                a_plugin.plugin_object.proc.terminate()

    def _remove_all_plugins(self, timeout=5):
        """
        Remove all plugins
        :param timeout: Maximum seconds to wait for subprocess to exit.
        :type timeout: ``int``
        """
        for input_plugin in self._input_plugins:
            a_plugin = self.manager.getPluginByName(input_plugin, 'Input')
            self._exit_plugin(a_plugin)
        for output_plugin in self._output_plugins:
            a_plugin = self.manager.getPluginByName(output_plugin, 'Output')
            self._exit_plugin(a_plugin)
        # TODO Transform Plugins

        # At this point all configured plugin subprocesses should be exiting
        # on their own.  However, if there is any plugin subprocess that didn't
        # exit for any reason, we must terminate them so we don't hang the
        # NuoCA process at exit.
        NuoCA.kill_all_plugin_processes(self.manager, timeout)

    def start(self):
        """
        Startup NuoCA
        """
        self._create_plugin_manager()
        self._activate_configured_plugins()
        interval_sync = IntervalSync(interval=self._collection_interval,
                                     seed_ts=self._starttime)

        # Collection Interval Loop
        loop_count = 0
        while self._enabled:
            loop_count += 1
            collection_timestamp = interval_sync.wait_for_next_interval()
            self._collection_cycle(collection_timestamp * 1000)
            if self._self_test:
                if loop_count >= self._config.SELFTEST_LOOP_COUNT:
                    self._enabled = False

    def shutdown(self, timeout=5):
        """
        Shutdown NuoCA
        :param timeout: Maximum seconds to wait for subprocess to exit.
        :type timeout: ``int``
        """
        nuoca_log(logging.INFO, "nuoca server shutdown")
        self._shutdown_all_plugins()
        self._remove_all_plugins(timeout)
        nuoca_logging_shutdown()


def signal_term_handler(signal, frame):
    global nuoca_obj
    if nuoca_obj:
        print "Got SIGTERM"
        sys.exit(0)


def nuoca_run(config_file, collection_interval, plugin_dir,
              starttime, mode, verbose, self_test,
              log_level, output_values):
    global nuoca_obj
    try:
        signal.signal(signal.SIGTERM, signal_term_handler)
        nuoca_obj = NuoCA(config_file, collection_interval, plugin_dir,
                          starttime, mode, verbose, self_test,
                          logging.getLevelName(log_level), output_values)
        nuoca_obj.start()
    except AttributeError as e:
        msg = str(e)
        nuoca_log(logging.ERROR, msg)
        print(msg)
    except Exception as e:
        msg = "Unhandled exception: %s" % e
        nuoca_log(logging.ERROR, msg)
        print(msg)
        stacktrace = traceback.format_exc()
        print(stacktrace)
    finally:
        if nuoca_obj:
            nuoca_obj.shutdown(nuoca_obj.config.SUBPROCESS_EXIT_TIMEOUT)
        print("Done.")


@click.command(help="CONFIG_FILE    path to NuoCA Collection Config file")
@click.argument('config-file', type=str)
@click.option('--collection-interval', default=30,
              help='Collection interval in seconds. Default:30')
@click.option('--plugin-dir', default=None,
              help='Path to NuoCA plugin directory')
@click.option('--starttime', default=None,
              help='Start time in Epoch seconds '
                   'for the first collection interval')
@click.option('--mode', default='coach',
              help='Run mode [coach | insights]')
@click.option('--verbose', is_flag=True, default=False,
              help='Run with verbose messages written to stdout')
@click.option('--self-test', is_flag=True, default=False,
              help='Run SELFTEST_LOOP_COUNT (default 5) '
                   'collection intervals then exit')
@click.option('--log-level', default='INFO',
              type=click.Choice(['CRITICAL', 'ERROR', 'WARNING',
                                 'INFO', 'DEBUG']),
              help='Set log level during test execution.')
@click.option('--output-values',
              '-o',
              multiple=True,
              default=None,
              help='One or more output values as '
                   'key=value pairs separated by commas. Multiples allowed')
def nuoca(config_file, collection_interval, plugin_dir,
          starttime, mode, verbose, self_test, log_level, output_values):
    nuoca_run(config_file, collection_interval, plugin_dir,
              starttime, mode, verbose, self_test, log_level, output_values)


if __name__ == "__main__":
    nuoca(prog_name="nuoca")



# (C) Copyright NuoDB, Inc. 2017-2020
#
# This source code is licensed under the MIT license found in the LICENSE
# file in the root directory of this source tree.

import logging
import os
import time
from math import sqrt
from stat import ST_CTIME

from pynuoca.nuoca_plugin import NuocaMPInputPlugin
from pynuoca.nuoca_util import nuoca_log


class Dbt2Plugin(NuocaMPInputPlugin):
    def __init__(self, parent_pipe):
        super(Dbt2Plugin, self).__init__(parent_pipe, 'Dbt2')
        self._dbt2_log_dir = None

    def startup(self, config=None):
        try:
            self._dbt2_log_dir = None
            required_config_items = ['dbt2_log_dir']
            if not self.has_required_config_items(config, required_config_items):
                return False
            self._dbt2_log_dir = config['dbt2_log_dir']
            return True
        except Exception as e:
            nuoca_log(logging.ERROR, str(e))
        return False

    def shutdown(self):
        pass

    @staticmethod
    def read_chunk(parser, fh, up_to_time, should_keep):
        line = fh.readline()
        if line:
            try:
                line_time = float(line.split(',')[0])
            except:
                line_time = 0
        else:
            return
        while line_time <= up_to_time:
            if should_keep:
                parser.process_mix_line(line)
            line = fh.readline()
            if not line:
                break
            try:
                line_time = float(line.split(',')[0])
            except:
                line_time = 0

    def collect(self, collection_interval):
        rval = None
        try:
            dbt2_parser = Dbt2ResultsParser()
            # Find two most recent mix logs
            if not os.path.exists(self._dbt2_log_dir):
                nuoca_log(logging.WARNING,
                          "DBT2 log path not found (" + self._dbt2_log_dir + ").")
            else:
                entries = (os.path.join(self._dbt2_log_dir, fn) for fn in
                           os.listdir(self._dbt2_log_dir) if
                           fn.startswith('mix'))
                entries = ((os.stat(path), path) for path in entries)
                entries = list(sorted(((stat[ST_CTIME], path) for stat,
                                                                  path in entries)))
                if len(entries) < 2:
                    nuoca_log(logging.INFO, "DBT2: There are not two mix logs yet.")
                else:
                    most_recent = entries[-1][1]
                    second_recent = entries[-2][1]
                    end_time = time.time()
                    start_time = end_time - collection_interval

                    with open(second_recent, 'r') as mixlog:
                        self.read_chunk(dbt2_parser, mixlog, start_time, False)
                        self.read_chunk(dbt2_parser, mixlog, end_time, True)

                    with open(most_recent, 'r') as mixlog:
                        self.read_chunk(dbt2_parser, mixlog, start_time, False)
                        self.read_chunk(dbt2_parser, mixlog, end_time, True)

                    metrics = super(Dbt2Plugin, self).collect(collection_interval)
                    metrics = dbt2_parser.process_metrics(metrics)
                    if metrics is not None:
                        rval = [metrics]
        except Exception as e:
            nuoca_log(logging.ERROR, str(e))
        return rval


"""
Mostly copied from dbt2's dbt2-post-process
Removed concept of rampup and steady-state since this is for real-time
monitoring
Removed tracking all metrics except New Order for efficiency.
"""


class Dbt2ResultsParser:
    def __init__(self):
        self.start_time = 0
        self.end_time = 0
        self.transactions = 0
        self.rollbacks = 0
        self.errors = 0
        self.sum_latencies = 0
        self.sum_latencies_sq = 0
        self.parse_errors = 0

    def process_mix_line(self, line):
        try:
            line_values = line.split(',')
            line_timestamp = int(line_values[0])
            if len(line_values) < 2 or line_values[1] == 'TERMINATED':
                return
            else:
                line_trans_type = line_values[1]
            if len(line_values) != 5:
                # probably result of partial buffered line
                self.parse_errors += 1
                return

            if self.start_time == 0:
                self.start_time = line_timestamp

            line_trans_rc = line_values[2]
            if line_trans_type == 'n' and line_trans_rc in ['C', 'R', 'E']:
                self.transactions += 1
                if line_trans_rc == 'R':
                    self.rollbacks += 1
                elif line_trans_rc == 'E':
                    self.errors += 1
                    return

                line_lat = float(line_values[3])
                self.sum_latencies += line_lat
                self.sum_latencies_sq += line_lat * line_lat
                self.end_time = line_timestamp
        except:
            self.parse_errors += 1
            return

    def process_metrics(self, metrics):
        if self.parse_errors > 0:
            nuoca_log(logging.INFO,
                      "DBT2: process_metrics: there were " + self.parse_errors +
                      " parse errors reading the logs")
        duration = float(self.end_time - self.start_time)
        if self.transactions > 1 and duration > 0:
            lat_avg = self.sum_latencies / self.transactions
            metrics['notpm'] = float(self.transactions) * 60.0 / duration
            metrics['lat_avg'] = float(lat_avg)
            # Estimating 90th percentile from standard deviation
            stddev = sqrt((self.sum_latencies_sq / self.transactions) -
                          (lat_avg * lat_avg))
            metrics['lat_90th'] = float(lat_avg + stddev * 1.28)
            return metrics
        else:
            return None



# (C) Copyright NuoDB, Inc. 2017-2020
#
# This source code is licensed under the MIT license found in the LICENSE
# file in the root directory of this source tree.

import logging
import re

from pynuoca.nuoca_plugin import NuocaMPInputPlugin
from pynuoca.nuoca_util import nuoca_log, search_running_processes, \
    execute_command, coerce_numeric


# Zabbix plugin
#
# This plugin expects that a zabbix agent is running on the localhost and
# zabbix_get is installed on the localhost.  The 'ZBX' section of the NuoCA
# config is used to control this Zabbix plugin.  The system metrics collected
# by zabbix is controlled by the 'keys' section.  The key names are in the
# Zabbix key name format.  The 'autoDiscoverMonitors' boolean is used to
# automatically include file system mounts, devices, and network interfaces in
# the collected 'keys'.  The loading of 'autoDiscoverMonitors' only happens
# when this plugin is initialized.  Dynamic changes to file system mounts,
# devices, or network interfaces will require reloading this plugin.
#
# This was developed and tested using Zabbix 2.2.11-1 on Ubuntu.
#
# Example Zabbix plugin configuration:
#
# - ZBX:
#    description : Collect machine stats from Zabbix
#    server: localhost
#    autoDiscoverMonitors: true
#    keys:
#    - system.uptime
#    - system.cpu.intr
#    - vm.memory.size[available]
#    - system.cpu.switches
#    - system.cpu.util[, interrupt]
#    - system.boottime
#    - system.cpu.util[, idle]
#    - system.cpu.util[, system]
#    - system.cpu.util[, iowait]
#    - system.cpu.util[, nice]
#    - system.cpu.util[, user]
#    - system.cpu.util[, softirq]
#    - system.localtime
#    - system.cpu.util[, steal]
#    - system.users.num
#    - proc.num[]
#    - vm.memory.size[total]
#    - system.uname
#    - system.hostname
#    - kernel.maxproc
#    - kernel.maxfiles
#


class ZabbixPlugin(NuocaMPInputPlugin):
    def __init__(self, parent_pipe):
        super(ZabbixPlugin, self).__init__(parent_pipe, 'ZBX')
        self._config = None

    def _append_config_key(self, key):
        nuoca_log(logging.INFO, "ZBX plugin appending config key: %s" % key)
        self._config['keys'].append(key)

    def _auto_discover_monitors(self):
        devices = []
        mounts = []
        interfaces = []

        # discover file system mounts & devices
        for line in execute_command("mount")[1].split("\n"):
            fields = line.strip().split()
            if len(fields) > 0 and fields[0].startswith("/dev"):
                devices.append(fields[0].strip())
                mounts.append(fields[2].strip())

        # discover network interfaces
        for line in execute_command("ip a")[1].split("\n"):
            if re.match("^[0-9]+:", line):
                fields = line.strip().split()
                interfaces.append(fields[1][0:-1])

        for device in devices:
            for op in ["read", "write"]:
                self._append_config_key("vfs.dev.%s[%s]" % (op, device))
        for mount in mounts:
            for op in ["used", "free"]:
                self._append_config_key("vfs.fs.size[%s,%s]" % (mount, op))
        for interface in interfaces:
            for op in ["in", "out"]:
                self._append_config_key("net.if.%s[%s]" % (op, interface))

    def startup(self, config=None):
        try:
            self._config = config

            # Make sure that zabbix agent is running.
            zabbix_agentd = search_running_processes('zabbix_agentd')
            if not zabbix_agentd:
                nuoca_log(logging.ERROR, "Zabbix agent daemon is not running.")
                return False

            # Make sure that zabbix_get is installed.
            command = "which zabbix_get"
            (exit_code, stdout, stderr) = execute_command(command)
            if exit_code != 0 or 'zabbix_get' not in stdout:
                nuoca_log(logging.ERROR, "Cannot locate zabbix_get command.")
                return False

            # Validate the configuration.
            required_config_items = ['autoDiscoverMonitors', 'server', 'keys']
            if not self.has_required_config_items(config, required_config_items):
                return False

            if config['autoDiscoverMonitors']:
                self._auto_discover_monitors()

            nuoca_log(logging.INFO, "ZBX plugin config: %s" %
                      str(self._config))

            return True
        except Exception as e:
            nuoca_log(logging.ERROR, str(e))
            return False

    def shutdown(self):
        pass

    def collect(self, collection_interval):
        rval = []
        collected_values = super(ZabbixPlugin, self).collect(collection_interval)
        try:
            for key in self._config['keys']:
                command = "zabbix_get -s localhost -k %s" % key
                (exit_code, stdout, stderr) = execute_command(command)
                if exit_code != 0:
                    nuoca_log(logging.ERROR, "Problem with zabbix_get command.")
                    return False
                if stdout.strip() is 'ZBX_NOTSUPPORTED':
                    nuoca_log(logging.WARNING,
                              "zabbix_get on key '%s' returned ZBX_NOTSUPPORTED" % key)
                collected_values[key] = coerce_numeric(stdout.strip())
            rval.append(collected_values)
        except Exception as e:
            nuoca_log(logging.ERROR, str(e))
        return rval



# (C) Copyright NuoDB, Inc. 2017-2020
#
# This source code is licensed under the MIT license found in the LICENSE
# file in the root directory of this source tree.

import datetime
import os
import time
import uuid
import sys
import hashlib
import logging
import logging.handlers
import subprocess
from copy import deepcopy
from logging import Handler
from nuoca_config import NuocaConfig

SECONDS_PER_DAY = 3600 * 24
DEFAULT_SEED_TS = 931752000  # Default NuoCA Epoch Timestamp

# NuoCA Temp Dir
if not os.path.exists(NuocaConfig.NUOCA_TMPDIR):
    os.mkdir(NuocaConfig.NUOCA_TMPDIR)

nuoca_logger = None
nuoca_loghandler = None
nuoca_collect_loghandler = None
yapsy_logger = None
yapsy_loghandler = None
last_log_message = None
last_log_error_message = None

# Global top level directory
nuoca_topdir = None  # Top level directory for NuoCA
nuoca_sourcedir = None

class UTC(datetime.tzinfo):
    """UTC tzinfo"""

    def __init__(self):
        self._zero = datetime.timedelta(0)

    def utcoffset(self, dt):
        return self._zero

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return self._zero


class CollectHandler(Handler):
    def __init__(self):
        self._nuoca_log_collect_queue = []
        Handler.__init__(self)

    def emit(self, record):
        """
        Emit a record.

        If a formatter is specified, it is used to format the record.
        """
        try:
            collect_item = {
                'timestamp': int(record.created * 1000.0),
                'NuoCA.filename': record.filename,
                'NuoCA.funcName': record.funcName,
                'NuoCA.log_level': record.levelname,
                'NuoCA.lineno': record.lineno,
                'NuoCA.message': record.message
            }
            if record.levelno != 20:
                collect_item['NuoCA.pathname'] = record.pathname
                collect_item['NuoCA.process'] = record.process
                collect_item['NuoCA.processName'] = record.processName
                collect_item['NuoCA.thread'] = record.thread
                collect_item['NuoCA.threadName'] = record.threadName
            self._nuoca_log_collect_queue.append(deepcopy(collect_item))
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

    @property
    def nuoca_log_collect_queue(self):
        return self._nuoca_log_collect_queue

    def collect(self, collection_interval):
        rval = None
        self._collection_interval = collection_interval
        try:
            collection_count = len(self._nuoca_log_collect_queue)
            if not collection_count:
                return rval

            rval = []
            for i in range(collection_count):
                collected_dict = self._nuoca_log_collect_queue.pop(0)
                rval.append(collected_dict)
        except Exception as e:
            pass
        return rval


def get_nuoca_collect_loghandler():
    global nuoca_collect_loghandler
    return nuoca_collect_loghandler


def initialize_logger(nuoca_logfile_name, collect_self=False):
    global nuoca_logger, nuoca_loghandler, nuoca_collect_loghandler
    global yapsy_logger, yapsy_loghandler

    logging.basicConfig(level=logging.INFO)
    # Global NuoCA logger
    nuoca_logger = logging.getLogger('nuoca')
    nuoca_loghandler = logging.handlers.WatchedFileHandler(nuoca_logfile_name)
    nuoca_loghandler.setLevel(logging.INFO)
    nuoca_loghandler.setFormatter(
        logging.Formatter('%(asctime)s NuoCA %(levelname)s %(message)s'))
    nuoca_logger.addHandler(nuoca_loghandler)

    # NuoCA loghandler to turn NuoCA logging events into a NuoCA collected item.
    if collect_self:
        nuoca_collect_loghandler = CollectHandler()
        nuoca_collect_loghandler.setLevel(logging.INFO)
        nuoca_collect_loghandler.setFormatter(
            logging.Formatter('%(asctime)s NuoCA %(levelname)s %(message)s'))
        nuoca_logger.addHandler(nuoca_collect_loghandler)

    # Global Yapsy logger
    yapsy_logger = logging.getLogger('yapsy')
    yapsy_loghandler = logging.FileHandler(nuoca_logfile_name)
    yapsy_loghandler.setLevel(logging.INFO)
    yapsy_loghandler.setFormatter(
        logging.Formatter('%(asctime)s YAPSY %(levelname)s %(message)s'))
    yapsy_logger.addHandler(yapsy_loghandler)
    if collect_self:
        yapsy_logger.addHandler(nuoca_collect_loghandler)


def randomid():
    """
    Returns a unique 32 character string for correlating different log lines.

    :returns str
    """
    # Visually easier in the logs to have a 32 char string without the dashes
    return hashlib.md5(str(uuid.uuid4())).hexdigest()


def function_exists(module, function):
    """
    Check Python module for specificed function.

    :param module: Python module
    :type module: ``types.ModuleType.``

    :param function: Name of the Python function
    :type ``str``

    :return: ``bool``
    """
    import inspect
    return hasattr(module, function) and any(
        function in f for f, _ in inspect.getmembers(
            module, inspect.isroutine))


def resolve_function(module, function):
    """
    Locate specified Python function in the specified Python package.

    :param module: A Python module
    :type module: ``types.ModuleType.``

    :param function: Name of Python function
    :type ``str``

    :return: Function or None if not found.
    """
    func = None
    if function_exists(module, function):
        func = getattr(module, function)
    if not func:
        nuoca_log(logging.ERROR, "Cannot find Python function %s in module %s" % (
            function, module
        ))
    return func


def get_nuoca_topdir():
    """
    Get the NuoCA top level directory

    :return: full path to the NuoCA top level directory.
    :type: str
    """
    global nuoca_topdir
    if not nuoca_topdir:
        this_file = os.path.realpath(__file__)
        this_dir = os.path.dirname(this_file)
        nuoca_topdir = os.path.abspath(os.path.join(this_dir, '..'))
    return nuoca_topdir


def get_nuoca_sourcedir():
    """
    Get the NuoCA python source directory

    :return: full path to the NuoCA python source directory
    :type: str
    """
    global nuoca_sourcedir
    if not nuoca_sourcedir:
        this_file = os.path.realpath(__file__)
        this_dir = os.path.dirname(this_file)
        nuoca_sourcedir = os.path.abspath(this_dir)
    return nuoca_sourcedir


def nuoca_set_log_level(log_level):
    """
    Set logging level
    :param log_level: logger log level
    :type: logger level
    """
    global nuoca_loghandler, yapsy_loghandler
    logging.getLogger('nuoca').setLevel(level=log_level)
    logging.getLogger('yapsy').setLevel(level=log_level)
    nuoca_loghandler.setLevel(log_level)
    if nuoca_collect_loghandler:
        nuoca_collect_loghandler.setLevel(log_level)
    yapsy_loghandler.setLevel(log_level)


def nuoca_log(log_level, msg):
    """
    Logger message
    :param log_level: logger log level
    :param msg: str: log message
    """
    global nuoca_logger, nuoca_loghandler
    global last_log_message, last_log_error_message

    if not msg:
        return

    last_log_message = msg
    if log_level == logging.ERROR:
        last_log_error_message = msg
    if not nuoca_logger:
        if msg[:-1] != '\n':
            msg = msg + '\n'
        sys.stderr.write(msg)
        return
    nuoca_logger.log(log_level, msg)
    nuoca_loghandler.flush()


def nuoca_logging_shutdown():
    """
    Shutdown ALL logging
    """
    logging.shutdown()


def nuoca_get_last_log_message():
    """
    Return last log message
    """
    return last_log_message


def nuoca_get_last_log_error_message():
    """
    Return last log error message
    """
    return last_log_error_message


def nuoca_gettimestamp():
    """
    Get the current Epoch time (Unix Timestamp)
    """
    return int(time.time())


def parse_keyval_list(options):
    """
    Convert list of key/value pairs (typically command line args) to a dict.

    Typically, each element in the list is of the form
        option=value
    However, multiple values may be specified in list elements by separating
    them with a comma (,) as in
        a=1,b=5,c=elbow
    Empty and all whitespace elements are also ignored

    :param options: a list of option values to parse
    :type options: list of str

    :return: dictionary of individual converted options
    :rtype: dict
    """

    ret = {}
    if not options:
        return ret  # ie empty dict

    # Convert -o options to a dict.  -o can be specified multiple times and can
    # have multiple values
    for opt in options:
        opt = opt.strip()
        for elem in opt.split(','):
            if not elem or elem.isspace():
                continue
            if '=' not in elem:
                raise AttributeError("key/value pair {} missing '='".format(elem))
            (k, v) = elem.split('=')
            k = k.strip()
            v = v.strip()
            ret[k] = v
    return ret


def search_running_processes(search_str):
    '''
    Return True if the search_str is in the command of any running process

    :param process_name: Name of process
    :type process_name: ``str``

    :return: True if process name is found, otherwise False
    '''
    try:
        p = subprocess.Popen(["ps", "axo", "comm"], stdout=subprocess.PIPE)
        out, err = p.communicate()
        if (search_str in out):
            return True
    except Exception:
        return False
    return False


def execute_command(command):
    '''
    Execute a posix command

    :param command: command line to execute
    :type command: ``str``

    :return: Python tuple of (exit_code, stdout, stderr)
    '''
    p = subprocess.Popen([command], shell=True, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()
    exit_code = p.returncode
    return (exit_code, stdout, stderr)


def coerce_numeric(s):
    '''
    Convert the string to an integer or float, if it is numeric.
    :param s:
    :type s: ``str``

    :return: integer, or float, or just a string.
    '''
    try:
        return int(s)
    except ValueError:
        try:
            return float(s)
        except ValueError:
            return s


class IntervalSync(object):
    """
    IntervalSync

    A universal temporal algorithm that can be used to automatically
    synchronize identical collection intervals across processes or nodes,
    without a predetermined starting timestamp.

    This can be used from cross node synchronization, assuming that the clock
    on all nodes are suitably synchronized (i.e. NTP time synchronization)
    and all nodes use the identical 'interval'.

    There is an optional seed timestamp.  If used, it must be identical for
    all processes/nodes that want to synchronize their intervals.  Clients
    that want to start the collections at specific time, should use seed_ts.

    """

    def __init__(self, interval, seed_ts=None):
        """
        Initialize IntervalSync

        :param interval: time in seconds
         :type ``int``
        :param seed_ts: Optional seed timestamp in UTC epoch seconds.  The
          seed_ts can be a point in the past, or a point in the future.  When
          seed_ts is a point in the future, the next interval will begin at that
          seed_ts.
         :type ``int``
        """
        self._interval = interval
        self._utc_tzinfo = UTC()
        self._seed_ts = seed_ts
        self._unix_epoch = datetime.datetime(1970, 1, 1, tzinfo=self._utc_tzinfo)

        if not self._seed_ts:
            self._seed_ts = DEFAULT_SEED_TS  # default to NuoCA Collection Epoch
        self._seed_dt = datetime.datetime.fromtimestamp(self._seed_ts,
                                                        self._utc_tzinfo)

    def compute_next_interval(self):
        """
        Compute the next time interval.

        :return: datetime object for the next interval.
        """
        now_dt = datetime.datetime.now(self._utc_tzinfo)
        seed_delta = now_dt - self._seed_dt
        seed_delta_total_seconds = seed_delta.total_seconds()
        if seed_delta_total_seconds >= 0:
            elasped_intervals = int(seed_delta_total_seconds) / self._interval
            next_interval = elasped_intervals + 1
            next_interval_seconds_from_seed = next_interval * self._interval
            next_interval_days_delta_from_seed = \
                next_interval_seconds_from_seed / SECONDS_PER_DAY
            next_interval_seconds_delta_from_seed = \
                next_interval_seconds_from_seed - \
                (next_interval_days_delta_from_seed * SECONDS_PER_DAY)
            next_interval_delta = datetime.timedelta(
                days=next_interval_days_delta_from_seed,
                seconds=next_interval_seconds_delta_from_seed)
            next_interval_dt = self._seed_dt + next_interval_delta
        else:
            next_interval_dt = self._seed_dt
        return next_interval_dt

    def wait_for_next_interval(self):
        """
        Wait for the next interval by sleeping until a precise point in time.

        Return UTC timestamp of the interval
        """
        next_interval_dt = self.compute_next_interval()

        # datetime.total_seconds() returns a floating point number
        # with microsecond accuracy.

        next_ts = int((next_interval_dt - self._unix_epoch).total_seconds())
        while True:
            utc_now = datetime.datetime.now(self._utc_tzinfo)
            diff = (next_interval_dt - utc_now).total_seconds()
            if diff <= 0:
                return next_ts
            time.sleep(diff / 2.0)


def to_bool(value):
    valid = {'true': True, 't': True, '1': True,
             'false': False, 'f': False, '0': False,
             }

    if isinstance(value, bool):
        return value

    if not isinstance(value, basestring):
        raise ValueError('invalid literal for boolean. Not a string.')

    lower_value = value.lower()
    if lower_value in valid:
        return valid[lower_value]
    else:
        raise ValueError('invalid literal for boolean: "%s"' % value)



# (C) Copyright NuoDB, Inc. 2017-2020
#
# This source code is licensed under the MIT license found in the LICENSE
# file in the root directory of this source tree.

import logging
from pynuoca.nuoca_plugin import NuocaMPInputPlugin
from pynuoca.nuoca_util import nuoca_log, coerce_numeric
import subprocess
import re
import os


class OLTPBenchPlugin(NuocaMPInputPlugin):
    def __init__(self, parent_pipe):
        super(OLTPBenchPlugin, self).__init__(parent_pipe, 'OLTPBench')
        self._config = None
        self._last_tps = None
        self._last_ltc = None
        self._log_file = None
        self._log_dir = None
        self._metrics_dict = None

    def startup(self, config=None):
        nuoca_log(logging.INFO, "Setting up oltpbenchplugin..")
        try:
            self._config = config
            required_config_items = ['log_dir', 'log_file']
            if not self.has_required_config_items(config, required_config_items):
                return False
            for var in 'log_file', 'log_dir':
                exec ("self._" + var + " = str(config['" + var + "'])")
            return True
        except Exception as e:
            nuoca_log(logging.ERROR, str(e))
            return False

    def shutdown(self):
        pass

    def run_parser(self):
        full_logpath = os.path.join(self._log_dir, self._log_file)
        cmd = "printf 'tps=\"'; \
printf $(grep 'Throughput: ' {} | tail -1 | sed 's/.*Throughput: //; s/ Tps//')\\\"; \
printf ' latency_average=\"'; \
printf $(grep 'Latency Average: ' {} | tail -1 | sed 's/.*Latency Average: //; s/ ms//')\\\";".format(full_logpath,
                                                                                                      full_logpath)
        nuoca_log(logging.DEBUG, "Running command: %s" % (format(cmd)))
        try:
            output = subprocess.check_output(
                cmd, stderr=subprocess.STDOUT, shell=True)
        except subprocess.CalledProcessError as e:
            nuoca_log(logging.ERROR, "run_parser ERROR: RC: %d, ERR: %s" %
                      (e.returncode, e.output))
            return None
        nuoca_log(logging.DEBUG, "Parser output: %s" % (format(output)))
        self._metrics_dict = dict(re.findall(r'(\w+)="([^"]+)"', output))
        if self._metrics_dict["tps"] == self._last_tps:
            self._metrics_dict["tps"] = None
        else:
            self._last_tps = self._metrics_dict["tps"]
        if self._metrics_dict["latency_average"] == self._last_ltc:
            self._metrics_dict["latency_average"] = None
        else:
            self._last_ltc = self._metrics_dict["latency_average"]

        return True

    def collect(self, collection_interval):
        rval = None
        try:
            nuoca_log(logging.DEBUG,
                      "Called collect() in OLTPBenchPlugin process...")
            collected_values = \
                super(OLTPBenchPlugin, self).collect(collection_interval)
            rval = []
            rval.append(collected_values)
            if not self.run_parser():
                return None
            if self._metrics_dict:
                for metric_item in self._metrics_dict:
                    if self._metrics_dict[metric_item]:
                        collected_values[metric_item] = \
                            coerce_numeric(self._metrics_dict[metric_item])
        except Exception as e:
            nuoca_log(logging.ERROR, str(e))
        return rval



# (C) Copyright NuoDB, Inc. 2017-2020
#
# This source code is licensed under the MIT license found in the LICENSE
# file in the root directory of this source tree.

from wrapt import decorator
import traceback, logging


@decorator
def print_exc(wrapped, instance, args, kwargs):
    try:
        return wrapped(*args, **kwargs)
    except Exception as e:
        print str(e)
        traceback.print_exc()


@decorator
def trace(wrapped, instance, args, kwargs):
    """ work on this """
    logging.getLogger('trace').info(
        "%s" % (wrapped.__qualname__ if hasattr(wrapped, '__qualname__') else wrapped.__name__))
    return wrapped(*args, **kwargs)



# (C) Copyright NuoDB, Inc. 2017-2020
#
# This source code is licensed under the MIT license found in the LICENSE
# file in the root directory of this source tree.

from nuomon_monitor import MetricsListener, EventListener
from nuomon_util import *
from pubsub import pub
import threading

from pynuoca.nuoca_util import nuoca_log

__all__ = ['MetricsConsumer', 'MetricsProducer', 'EventConsumer', 'EventProducer']


class AtomicCounter:
    """An atomic, thread-safe incrementing counter."""

    def __init__(self, initial=0):
        """Initialize a new atomic counter to given initial value (default 0)."""
        self.value = initial
        self._lock = threading.Lock()

    def increment(self, num=1):
        """Atomically increment the counter by num (default 1) and return the new value. """
        with self._lock:
            self.value += num
            return self.value


# Need request/reply for current state.
class EventConsumer(object):
    id_generator = AtomicCounter(0)

    def __init__(self, **args):
        for k, v in args.iteritems():
            setattr(self, k, v)
        replyToId = str(EventConsumer.id_generator.increment())
        self.__replyChannel = 'event.request.%s' % (replyToId)

    def __request_latest(self):
        pub.subscribe(self._onEvent, self.__replyChannel)
        pub.sendMessage('event.sync.request', replyTo=self.__replyChannel)
        pub.unsubscribe(self._onEvent, self.__replyChannel)

    def request_domain_state(self):
        self.__request_latest()

    @print_exc
    def start(self):
        # request full set of data to send then subscribe for data stream.
        self.__request_latest()
        pub.subscribe(self._onEvent, 'event.queue')

    @print_exc
    def stop(self):
        print '%s.stop' % (self.__class__.__name__)
        pub.unsubscribe(self._onEvent, 'event.queue')

    def onEvent(self, event, entity_type, entity_data, event_data):
        pass

    def _onEvent(self, event, entity_type, entity_data, event_data):
        self.onEvent(event, entity_type, entity_data, event_data)


class MetricsConsumer(object):
    """
    baseclass to listen to and request metric events
    Two types of events:
    -  description - describes each metric
    -  values      - update values or latest values (on update)
    """
    id_generator = AtomicCounter(0)

    def __init__(self, **args):
        """ args (that this class will look at)
            mode :  "full", "changed"
            filter : <callable>(identity)
        """
        for k, v in args.iteritems():
            setattr(self, k, v)
        self.mode = getattr(self, 'mode', 'full')
        self.filter = getattr(self, 'filter', lambda x: True)

        self.replyToId = str(MetricsConsumer.id_generator.increment())
        self.replyMetricsChannel = 'metrics.requested.description.%s' % (self.replyToId)
        self.replyValuesChannel = 'metrics.requested.values.%s' % (self.replyToId)
        pass

    def __del__(self):
        pass

    @property
    def instanceId(self):
        return self.replyToId

    def onMetrics(self, description):
        pass

    def onValues(self, values):
        pass

    def __request_latest(self):
        pub.subscribe(self._onMetrics, self.replyMetricsChannel)
        pub.subscribe(self._onValues, self.replyValuesChannel)
        pub.sendMessage('metrics.request.description', replyTo=self.replyMetricsChannel)
        pub.sendMessage('metrics.request.values', replyTo=self.replyValuesChannel)
        pub.unsubscribe(self._onMetrics, self.replyMetricsChannel)
        pub.unsubscribe(self._onValues, self.replyValuesChannel)

    def request_metrics(self):
        self.__request_latest()

    @print_exc
    def start(self):
        # request full set of data to send then subscribe for data stream.
        self.__request_latest()
        if self.mode == 'full':
            pub.subscribe(self._onValues, 'metrics.values.full')
        else:
            pub.subscribe(self._onValues, 'metrics.values.updated')

    @print_exc
    def stop(self):
        print '%s.stop' % (self.__class__.__name__)
        if self.mode == 'full':
            pub.unsubscribe(self._onValues, 'metrics.values.full')
        else:
            pub.unsubscribe(self._onValues, 'metrics.values.updated')
        pub.subscribe(self._onShutdown, self.replyValuesChannel)
        pub.sendMessage('metrics.request.values', replyTo=self.replyValuesChannel)
        pub.unsubscribe(self._onShutdown, self.replyValuesChannel)

    def _filter(self, identity):
        return self.filter(identity)
        # filter = getattr(self,'filter',None)
        # return filter is None or filter(identity)

    @print_exc
    @trace
    def _onMetrics(self, description):
        self.onMetrics(description)
        pass

    @print_exc
    @trace
    def _onValues(self, identity, body):
        # publish 'full' latest values, 'changed' modified values
        if self._filter(identity):
            body.update(identity.toValues())
            self.onValues(body)

    @print_exc
    @trace
    def _onShutdown(self, identity, body):
        """ publish one more set of values - all zeros """
        print '%s._onShutdown' % (self.__class__.__name__)
        if not self._filter(identity):
            return
        if self.mode != 'full':
            # changed mode we only want to publish values != 0
            removekeys = [k for k, v in body.iteritems() if v == 0]
            for k in removekeys:
                del body[k]
        zerokeys = [k for k, v in body.iteritems() if v != 0 and type(v) is int]
        for k in zerokeys:
            body[k] = 0
        body.update(identity.toValues())
        print '%s._onShutdown: %s' % (self.__class__.__name__, body)
        self.onValues(body)


class EventProducer(EventListener):
    def __init__(self):
        super(EventListener, self).__init__()
        self.cache = {EventListener.EntityType.PEER: [],
                      EventListener.EntityType.PROCESS: [],
                      EventListener.EntityType.DATABASE: []
                      }
        pub.subscribe(self.onEventRequest, 'event.sync.request')

    def onEvent(self, event, entity_type, entity_data, event_data):
        if entity_type in self.cache:
            if event == EventListener.EventType.JOINED:
                cached = self.cache[entity_type]
                cached.append(entity_data)
            elif event == EventListener.EventType.LEFT:
                cached = self.cache[entity_type]
                for obj, ndx in enumerate(cached):
                    if obj == entity_data:
                        del obj[ndx]
                        break
        pub.sendMessage('event.queue', event=event, entity_type=entity_type, entity_data=entity_data,
                        event_data=event_data)
        pass

    def onEventRequest(self, replyTo):
        peers = self.cache[EventListener.EntityType.PEER]
        for edata in peers:
            pub.sendMessage(replyTo, event=EventListener.EventType.JOINED, entity_type=EventListener.EntityType.PEER,
                            entity_data=edata, event_data=None)
        dbs = self.cache[EventListener.EntityType.DATABASE]
        for edata in dbs:
            pub.sendMessage(replyTo, event=EventListener.EventType.JOINED,
                            entity_type=EventListener.EntityType.DATABASE,
                            entity_data=edata, event_data=None)
        processes = self.cache[EventListener.EntityType.PROCESS]
        for edata in processes:
            pub.sendMessage(replyTo, event=EventListener.EventType.JOINED, entity_type=EventListener.EntityType.PROCESS,
                            entity_data=edata, event_data=None)


class MetricsProducer(MetricsListener):
    def __init__(self):
        super(MetricsProducer, self).__init__()
        nuoca_log(logging.DEBUG, "nuomon_broadcast MetricsProducer.__init__()")
        self.identity = None

    @print_exc
    def onStart(self, metrics):
        super(MetricsProducer, self).onStart(metrics)
        nuoca_log(logging.DEBUG, "nuomon_broadcast MetricsProducer.onStart()")
        pub.sendMessage('metrics.description', description=self.metrics)
        pub.subscribe(self.onMetricsRequest, 'metrics.request.description')
        pub.subscribe(self.onValuesRequest, 'metrics.request.values')

    @print_exc
    def onChange(self, values):
        nuoca_log(logging.DEBUG, "nuomon_broadcast MetricsProducer.onChange()")
        if len(self.values) > 0:
            pub.sendMessage('metrics.values.updated', identity=self.getIdentity(), body=values)
        super(MetricsProducer, self).onChange(values)
        pub.sendMessage('metrics.values.full', identity=self.getIdentity(), body=self.values)

    @print_exc
    @trace
    def onMetricsRequest(self, replyTo):
        pub.sendMessage(replyTo, description=self.metrics)

    @print_exc
    @trace
    def onValuesRequest(self, replyTo):
        if self.identity:
            pub.sendMessage(replyTo, identity=self.identity, body=self.values)

    def getIdentity(self):
        class Identity(object):
            def __init__(self, **kwds):
                self.__dict__ = kwds

            def toValues(self):
                return dict(Database=self.database,
                            NodeId=self.nodeid,
                            Hostname=self.hostname,
                            ProcessId=self.processid,
                            NodeType=self.nodetype)

        if self.identity is None:
            values = self.values
            self.identity = Identity(database=values['Database'],
                                     nodeid=values['NodeId'],
                                     hostname=values['Hostname'],
                                     processid=values['ProcessId'],
                                     nodetype=values['NodeType'])
        return self.identity


if __name__ == "__main__":
    from nuomon_monitor import get_nuodb_metrics
    import time, traceback, yaml, sys


    class Handle(MetricsConsumer):
        """ example handler that listens for messages from BradcastListener."""

        def __init__(self):
            super(Handle, self).__init__()
            pass

        def onMetrics(self, description):
            print yaml.dump(description)
            pass

        def onValues(self, values):
            print yaml.dump(values)
            pass


    broker = 'localhost'
    d = get_nuodb_metrics(broker, 'bird', listener=MetricsProducer)

    # request / reply assumes that data has already been cached.
    time.sleep(2)
    obj = Handle()

    print 'start...'
    obj.start()
    time.sleep(1)
    d.disconnect()
    # d.wait_forever()



#!/usr/bin/python

# (C) Copyright NuoDB, Inc. 2017-2020
#
# This source code is licensed under the MIT license found in the LICENSE
# file in the root directory of this source tree.

import os, sys, yaml
import time, aenum
import socket
import threading, logging
from nuomon_util import *

from pynuodb.session import Session, BaseListener, SessionMonitor
from pynuodb.entity import Domain
from xml.etree import ElementTree

# base class, do not override methods

__all__ = ['MetricsListener', 'EventListener', 'get_nuodb_metrics']


class BaseMetricsListener(BaseListener):
    def __init__(self):
        super(BaseMetricsListener, self).__init__()
        self.__first = True
        self.__process = None

    @property
    def process(self):
        return self.__process

    @process.setter
    def process(self, p):
        self.__process = p

    def __get_item(self, attrs):
        units = ["COUNT", "MILLISECONDS", "STATE",
                 "NUMBER", "PERCENT", "IDENTIFIER",
                 "DELTA"]
        return {"unit": units[int(attrs['units']) - 1],
                "description": attrs['header']}

    # @trace
    def message_received(self, root):
        def parseStr(x):
            try:
                return int(x)
            except:
                return x

        items = {}
        if root.tag == "Items":
            for child in root:
                items[child.attrib['name']] = self.__get_item(child.attrib)
            items['Database'] = dict(unit="IDENTIFIER", description="Database Name")
            items['Region'] = dict(unit="IDENTIFIER", description="Region Name")
            items['DomainName'] = dict(unit="IDENTIFIER", description="Domain Name")
            items['Version'] = dict(unit="IDENTIFIER", description="NuoDB Version String")
            self.onStart(items)
        elif root.tag == 'Status':
            values = dict([(k, parseStr(v)) for k, v in root.attrib.iteritems()])
            if self.__first:
                values['Database'] = self.process.database.name
                values['Region'] = self.process.peer.get_tag('region')
                values['DomainName'] = self.process.peer.domain.domain_name
                values['Version'] = self.process.version
                self.__first = False
            values['TimeStamp'] = int(time.time() * 1000.0)
            self.onChange(values)

    def closed(self):
        self.onEnd()
        pass


# interface, do override
# onEvent
class EventListener(object):
    class EventType(int, aenum.Enum):
        JOINED = 1
        LEFT = -1
        CHANGED = 0
        FAILED = -2

    class EntityType(str, aenum.Enum):
        DATABASE = 'database'
        PEER = 'peer'
        PROCESS = 'process'

    def __init__(self):
        pass

    def onEvent(self, event, entity_type, entity_data, event_data):
        #        print 'onEvent',event, entity_type, entity_data
        #        if event_data is None:
        #            print event_data
        #        else:
        #            print event_data
        pass


# interface, do override
#   onStart
#   onStat
#   onEnd


class MetricsListener(BaseMetricsListener):
    """ Base class for metrics collection.
    Remembers previous values"""

    def __init__(self):
        super(MetricsListener, self).__init__()
        self.__metrics = {}
        self.__values = {}
        pass

    @property
    def metrics(self):
        return self.__metrics

    @property
    def values(self):
        return self.__values

    def init(self, args):
        return self

    def onStart(self, metrics):
        """ remembers metrics  """
        self.__metrics = metrics

    def onChange(self, metrics):
        """ remembers previous values """
        self.__values.update(metrics)

    def onEnd(self):
        """ zero all values """
        zeroMetrics = {}
        for k, v in self.__values.iteritems():
            if v != 0 and type(v) is int:
                zeroMetrics[k] = 0
        zeroMetrics['TimeStamp'] = int(time.time() * 1000.0)
        self.onChange(zeroMetrics)
        pass


class MetricsDomain(Domain):
    def __init__(self, broker, user, password, listener):
        Domain.__init__(self, broker, user, password, listener)

    def wait_forever(self, log=False):
        try:
            while True:
                time.sleep(10)
        except:
            pass
        finally:
            self.disconnect()
            if log:
                print "disconnect..."


# @print_exc
def get_nuodb_metrics(broker, password, listener, user='domain', database=None, host=None, process=None, args=None,
                      domain_listener=None):
    # Listener is class derived from MetricsListener
    if listener is None:
        listener = MetricsListener
    if domain_listener is None:
        domain_listener = EventListener

    domain = DomainListener(user=user,
                            password=password,
                            database=database,
                            host=host,
                            process=process,
                            listener=listener,
                            args=args,
                            domain_listener=domain_listener())
    return MetricsDomain(broker, user, password, domain)


# implementation
""" TODO:  Add event stream """


class DomainListener(object):
    cached_addresses = {}
    running_monitors = {}

    def __init__(self, **kwds):
        for k, v in kwds.iteritems():
            setattr(self, k, v)
        self.listener = getattr(self, 'listener', MetricsListener)
        self.domain_listener = getattr(self, 'domain_listener', EventListener())

    def __monitoring_peer(self, peer):
        """ apply filters if specified """
        return True

    def peer_joined(self, peer):
        # apply filters if specified
        if not self.__monitoring_peer(peer):
            return
        id = dict(hostname=peer.hostname,
                  address=peer.address,
                  port=peer.port,
                  # is_broker= peer.broker,
                  # agent_id=  peer.agent_id,
                  version=peer.version)
        self.domain_listener.onEvent(EventListener.EventType.JOINED,
                                     EventListener.EntityType.PEER,
                                     id, None)

    def peer_left(self, peer):
        # apply filters if specified
        if not self.__monitoring_peer(peer):
            return
        id = dict(hostname=peer.hostname,
                  address=peer.address,
                  port=peer.port,
                  # is_broker= peer.broker,
                  # agent_id=  peer.agent_id,
                  version=peer.version)
        self.domain_listener.onEvent(EventListener.EventType.LEFT,
                                     EventListener.EntityType.PEER,
                                     id, None)

    def process_joined(self, p):
        logging.debug("process joined: %s" % str(p))
        db = p.database
        # apply filters if specified
        if self.database and self.database != db.name:
            return
        if self.host:
            addresses = [p.hostname, p.peer.address, p.peer.hostname,
                         self.getaddr(p.hostname),
                         self.getaddr(p.peer.address),
                         self.getaddr(p.peer.hostname)]
            if self.host not in addresses:
                return
        if self.process and self.database and self.process != p.node_id:
            return
        if self.process and self.host and self.process != p.pid:
            return

        # setup monitors
        id = dict(hostname=p.hostname,
                  dbname=p.database.name,
                  port=p.port,
                  pid=p.pid,
                  # is_transaction = p.transactional,
                  version=p.version,
                  node_id=p.node_id)
        self.domain_listener.onEvent(EventListener.EventType.JOINED,
                                     EventListener.EntityType.PROCESS,
                                     id, None)
        self.monitorEngine(p)

    def process_left(self, p):
        logging.debug("process left: %s" % str(p))
        if str(p) in self.running_monitors:
            id = dict(hostname=p.hostname,
                      dbname=p.database.name,
                      port=p.port,
                      pid=p.pid,
                      # is_transaction = p.transactional,
                      version=p.version,
                      node_id=p.node_id)
            self.domain_listener.onEvent(EventListener.EventType.LEFT,
                                         EventListener.EntityType.PROCESS,
                                         id, None)
            monitor = self.running_monitors[str(p)]
            del self.running_monitors[str(p)]

    def process_failed(self, peer, reason):
        print 'failed: ', peer
        id = dict(hostname=peer.hostname,
                  address=peer.address,
                  port=peer.port,
                  # is_broker= peer.broker,
                  # agent_id=  peer.agent_id,
                  version=peer.version)
        self.domain_listener.onEvent(EventListener.EventType.FAILED,
                                     EventListener.EntityType.PROCESS,
                                     id, reason)
        pass

    def process_status_changed(self, p, status):
        id = dict(hostname=p.hostname,
                  dbname=p.database.name,
                  port=p.port,
                  pid=p.pid,
                  # is_transaction = p.transactional,
                  version=p.version,
                  node_id=p.node_id)
        self.domain_listener.onEvent(EventListener.EventType.CHANGED,
                                     EventListener.EntityType.PROCESS,
                                     id, status)

    def database_joined(self, database):
        id = dict(database=database.name)
        self.domain_listener.onEvent(EventListener.EventType.JOINED,
                                     EventListener.EntityType.DATABASE,
                                     id, None)
        pass

    def database_left(self, database):
        id = dict(database=database.name)
        self.domain_listener.onEvent(EventListener.EventType.LEFT,
                                     EventListener.EntityType.DATABASE,
                                     id, None)
        pass

    def getaddr(self, hostname):
        try:
            if hostname not in self.cache_addresses:
                self.cache_addressed[hostname] = socket.gethostbyname(hostname)
        except:
            self.cached_addresses[hostname] = None
        return self.cached_addresses[hostname]

    @print_exc
    def monitorEngine(self, process):
        """ Monitor statistics from a TE or SM """

        # attach and monitor stats from engine
        engine_key = self.__get_engine_key(process)
        engine_session = Session(process.address, port=process.port, service="Monitor")
        engine_session.authorize("Cloud", engine_key)

        args = getattr(self, 'args', None)
        if args is None:
            callbk = self.listener()
        else:
            callbk = self.listener().init(args)
        callbk.process = process
        monitor = SessionMonitor(engine_session, listener=callbk)
        monitor.start()
        engine_session.doConnect()
        self.running_monitors[str(process)] = monitor

    @print_exc
    def getSyncTrace(self, process):
        # attach and monitor stats from engine
        engine_key = self.__get_engine_key(process)
        engine_session = Session(process.address, port=process.port, service="Monitor")
        engine_session.authorize("Cloud", engine_key)

    def closed(self):
        for name, monitor in self.running_monitors.iteritems():
            monitor.close()
        self.running_monitors = {}

    def __get_engine_key(self, process):
        session = Session(process.peer.connect_str, service="Manager")
        session.authorize(self.user, self.password)
        pwd_response = session.doRequest(attributes={"Type": "GetDatabaseCredentials",
                                                     "Database": process.database.name})
        pwd_xml = ElementTree.fromstring(pwd_response)
        pwd = pwd_xml.find("Password").text.strip()
        return pwd

    def waitForTerminate(self):
        while len(self.running_monitors):
            threading.Thread.join(self.running_monitors.values()[0])


if __name__ == "__main__":
    import optparse

    parser = optparse.OptionParser(usage="%prog [options] [hostname:port]",
                                   description="""Attaches
                                   to nodes and recieves stat data on a periodic basis.  You can
                                   filter which processes to attach to.  For a single process specify via
                                   a given host and pid or database and nodeid.  For all processes on a
                                   host specify the host.  For all processes of a given database specify
                                   the database.
                                   """)
    parser.add_option("-u", "--user", dest="user", default="domain", help="Domain user (domain).")
    parser.add_option("-p", "--password", dest="password", default="bird", help="Domain password (bird).")
    parser.add_option("-d", "--database", dest="database", default=None, help="Monitor given database.")
    parser.add_option("-n", "--host", dest="host", default=None, help="Only monitor process on this host.")
    parser.add_option("-i", "--process", dest="process", default=None, type='int',
                      help="Process identifier, pid (with host) or nodeId (with datbase).")
    parser.add_option("-m", "--mode", dest="mode", choices=['full', 'changed'], default='changed',
                      help="Controls output.")
    (options, args) = parser.parse_args()

    # validate input 
    if options.process and options.database and options.host:
        parser.print_help()
        sys.exit(1)
    if options.process and not (options.database or options.host):
        parser.print_help()
        sys.exit(1)

    broker = "localhost"
    if len(args) == 1:
        broker = args[0]
    elif len(args) > 1:
        parser.print_help()
        sys.exit(1)


    class DemoListener(MetricsListener):
        DOHEADER = True

        def __init__(self):
            super(DemoListener, self).__init__()
            self.mode = 'full'

        def init(self, args):
            if 'mode' in args:
                self.mode = args['mode']
            return self

        def onStart(self, metrics):
            super(DemoListener, self).onStart(metrics)
            if DemoListener.DOHEADER:
                DemoListener.DOHEADER = False
                print "Description of metrics..."
                for k, v in self.desciption.iteritems():
                    print "%-48s: %s" % ("%s(%s)" % (k, v['unit']), v['description'])
                print

        @print_exc
        def onChange(self, values):
            super(DemoListener, self).onChange(values)
            if self.mode == 'full':
                use = self.values
            else:
                use = values
            header = "%s" % (self.process)
            display = {header: use}
            print yaml.dump(display, default_flow_style=False)

        def onEnd(self):
            super(DemoListener, self).onEnd()
            print "Goodbye from: %s:%s %s(%s)" % (
            self.values['Hostname'], self.values['ProcessId'], self.values['Database'], self.values['NodeId'])
            pass


    args = {'mode': options.mode}
    del options.mode
    d = get_nuodb_metrics(broker=broker, listener=DemoListener, args=args, **options.__dict__)
    d.wait_forever()



# Copyright (c) 2017-2020, NuoDB, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of NuoDB, Inc. nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUODB, INC. BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import logging
from pynuoca.nuoca_plugin import NuocaMPOutputPlugin
from pynuoca.nuoca_util import nuoca_log
import requests
from pynuoca.lib import metrics_influx
import gzip

try:
    from cStringIO import StringIO
except:
    from StringIO import StringIO
from distutils.util import strtobool
from requests.auth import HTTPBasicAuth


class InfluxDBPlugin(NuocaMPOutputPlugin):
    def __init__(self, parent_pipe, config=None):
        super(InfluxDBPlugin, self).__init__(parent_pipe, 'InfluxDB')
        self._config = config

    def startup(self, config=None):
        try:
            self._config = config
            cvalue = self._config.get('compress', 'False')
            self._compress = strtobool(str(cvalue))
            self._user = self._config.get('user', None)
            self._password = self._config.get('password', None)

            required_config_items = ['url']
            if not self.has_required_config_items(config, required_config_items):
                return False

            good = self._password is None if self._user is None else self._password is not None
            if not good:
                nuoca_log(logging.ERROR,
                          "%s plugin: if user specified in config, password must be also and vice versa" %
                          (self._plugin_name))
            return good
        except Exception as e:
            nuoca_log(logging.ERROR, str(e))
            return False

    def shutdown(self):
        pass

    def format(self, ts_values):
        pass

    def store(self, ts_values):

        rval = None
        try:
            nuoca_log(logging.DEBUG, "Called store() in NuoCA InfluxDBPlugin process")
            rval = super(InfluxDBPlugin, self).store(ts_values)

            # metrics_influx only formats data from NuoMon.
            # need to look at handling ZBX also.

            def write(fileobj):
                for single_values in ts_values:
                    values = dict([(k[7:], v) for k, v in single_values.iteritems() if k.startswith("NuoMon.")])
                    if len(values) == 0:
                        continue
                    contentType, message = metrics_influx.format(values)
                    fileobj.write(message)

            s = StringIO()
            if self._compress:
                with gzip.GzipFile(fileobj=s, mode='wb') as f:
                    write(f)
                headers = {"Content-Encoding": "gzip",
                           "Content-Type": "text/plain; charset=utf-8"
                           }
            else:
                write(s)
                headers = {"Content-Type": "text/plain; charset=utf-8"}

            mbuffer = s.getvalue()
            s.close()
            url = self._config['url']
            if self._user is not None:
                arg = requests.post(url, headers=headers, data=mbuffer, auth=HTTPBasicAuth(self._user, self._password))
            else:
                arg = requests.post(url, headers=headers, data=mbuffer)

            if arg.status_code not in [200, 204]:
                nuoca_log(logging.ERROR, "%d : %s" % (arg.status_code, arg.text))
        except Exception as e:
            nuoca_log(logging.ERROR, str(e))
        return rval



# (C) Copyright NuoDB, Inc. 2017-2020
#
# This source code is licensed under the MIT license found in the LICENSE
# file in the root directory of this source tree.

import os
import logging
from elasticsearch import Elasticsearch
from elasticsearch import helpers
from pynuoca.nuoca_plugin import NuocaMPOutputPlugin
from pynuoca.nuoca_util import nuoca_log


class ElasticSearchPlugin(NuocaMPOutputPlugin):
    def __init__(self, parent_pipe, config=None):
        super(ElasticSearchPlugin, self).__init__(parent_pipe, 'ElasticSearch')
        self._config = config
        self.elastic_hosts = None
        self.port = None
        self.url_prefix = None
        self.es_obj = None
        self.es_index = None
        self.es_index_pipeline = None
        self.ES_LIMIT_SIZE = 1000
        self.ES_BULK_CHUNK_SIZE = 75
        self.ES_BULK_CHUNK_BYTES = 6291456

    def startup(self, config=None):
        try:
            logger = logging.getLogger('elasticsearch')
            logger.setLevel(logging.WARNING)
            self._config = config
            required_config_items = ['HOST', 'INDEX']
            if not self.has_required_config_items(config, required_config_items):
                return False
            if 'PIPELINE' in config:
                self.es_index_pipeline = os.path.expandvars(config['PIPELINE'])
            host = os.path.expandvars(self._config['HOST'])
            if 'PORT' in self._config:
                if isinstance(self._config['PORT'], int):
                    self.port = self._config['PORT']
                else:
                    self.port = os.path.expandvars(self._config['PORT'])
            if 'URL_PREFIX' in config:
                self.url_prefix = os.path.expandvars(self._config['URL_PREFIX'])
            self.es_index = os.path.expandvars(self._config['INDEX'])
            elastic_host = {'host': host}
            if self.port:
                elastic_host['port'] = self.port
            if self.url_prefix:
                elastic_host['url_prefix'] = self.url_prefix
            self.elastic_hosts = [elastic_host]
            self.es_obj = Elasticsearch(self.elastic_hosts, timeout=10)
            return True
        except Exception as e:
            nuoca_log(logging.ERROR, "ElasticSearch Store error: %s" % str(e))
            return False

    def shutdown(self):
        pass

    def _bulk_index(self, ts_values):
        offset = 0
        total = 0
        b_complete = False
        limit_size = self.ES_LIMIT_SIZE
        update_count = 0
        total = len(ts_values)
        es_index = self.es_index
        if self.es_index_pipeline:
            es_index
        while not b_complete:
            actions = []
            items_to_process = limit_size
            if offset + limit_size > total:
                items_to_process = total - offset
            for i in range(items_to_process):
                row = ts_values[i]
                action = {"_index": es_index,
                          "_type": "nuoca",
                          "_source": row}
                actions.append(action)
            if len(actions) > 0:
                resp = helpers.bulk(self.es_obj,
                                    actions,
                                    chunk_size=self.ES_BULK_CHUNK_SIZE,
                                    max_chunk_bytes=self.ES_BULK_CHUNK_BYTES)
                nuoca_log(logging.DEBUG, "ES bluk response: %s" % str(resp))
                update_count += resp[0]
            if offset + items_to_process >= total:
                b_complete = True
            offset += limit_size
        nuoca_log(logging.DEBUG, "ES Indexed %d documents" % update_count)

    def store(self, ts_values):
        rval = None
        try:
            nuoca_log(logging.DEBUG, "Called store() in NuoCA ElasticSearch process")
            rval = super(ElasticSearchPlugin, self).store(ts_values)
            self._bulk_index(ts_values)
        except Exception as e:
            nuoca_log(logging.ERROR, str(e))
        return rval



# (C) Copyright NuoDB, Inc. 2017-2020
#
# This source code is licensed under the MIT license found in the LICENSE
# file in the root directory of this source tree.

import logging
from pynuoca.nuoca_plugin import NuocaMPOutputPlugin
from pynuoca.nuoca_util import nuoca_log


class PrinterPlugin(NuocaMPOutputPlugin):
    def __init__(self, parent_pipe, config=None):
        super(PrinterPlugin, self).__init__(parent_pipe, 'Printer')
        self._config = config

    def startup(self, config=None):
        try:
            self._config = config
            return True
        except Exception as e:
            nuoca_log(logging.ERROR, str(e))
            return False

    def shutdown(self):
        pass

    def store(self, ts_values):
        rval = None
        try:
            nuoca_log(logging.DEBUG, "Called store() in MPCounterPlugin process")
            rval = super(PrinterPlugin, self).store(ts_values)
            print(ts_values)
        except Exception as e:
            nuoca_log(logging.ERROR, str(e))
        return rval



# (C) Copyright NuoDB, Inc. 2017-2020
#
# This source code is licensed under the MIT license found in the LICENSE
# file in the root directory of this source tree.

import logging
import os
import re
import socket
import threading
import time

from copy import deepcopy
from pynuoca.nuoca_plugin import NuocaMPInputPlugin
from pynuoca.nuoca_util import nuoca_log
from pynuoca.nuomon.nuomon_monitor import DomainListener, MetricsDomain
from pynuoca.nuomon.nuomon_broadcast import MetricsConsumer, MetricsProducer, \
    MetricsListener, EventListener


# NuoMonitor plugin
#
# Example NuoMonitor plugin configuration:
#
# - NuoMonitor:
#    description : Collection from internal nuomonitor tool
#    database_regex_pattern: dbt2
#    broker: 172.19.0.16
#    domain_username: domain
#    domain_password: bird


class NuoMonHandler(MetricsConsumer):
    """ NuoMon handler that listens for messages from BroadcastListener."""

    nuo_monitor_obj = None

    def __init__(self, nuo_monitor_obj):
        super(NuoMonHandler, self).__init__()
        nuoca_log(logging.DEBUG, "NuoMon: NuoMonHandler.__init__()")
        self.nuo_monitor_obj = nuo_monitor_obj
        pass

    def onMetrics(self, description):
        nuoca_log(logging.DEBUG, "NuoMon: NuoMonHandler.onMetrics()")
        pass

    def onValues(self, values):
        nuoca_log(logging.DEBUG, "NuoMon: NuoMonHandler.onValues()")
        self.nuo_monitor_obj.nuomonitor_collect_queue.append(deepcopy(values))
        pass


class NuoMonitorPlugin(NuocaMPInputPlugin):
    def __init__(self, parent_pipe):
        super(NuoMonitorPlugin, self).__init__(parent_pipe, 'NuoMon')
        self._config = None
        self._broker = None
        self._enabled = False
        self._numon_handler_ready = False
        self._domain = None
        self._domain_username = 'domain'
        self._domain_password = 'bird'
        self._domain_metrics = None
        self._domain_metrics_host = None
        self._database_regex_pattern = '.*'
        self._host_uuid_shortname = False
        self._thread = None
        self._nuomonitor_collect_queue = []
        self._collection_interval = 30

    def wait_for_terminate(self):
        if self._domain:
            self._domain.waitForTerminate()

    def get_nuodb_metrics(self, broker, password, listener, user='domain',
                          database=None, host=None, process=None, args=None,
                          domain_listener=None):
        # Listener is class derived from MetricsListener
        if listener is None:
            listener = MetricsListener
        if domain_listener is None:
            domain_listener = EventListener

        self._domain = DomainListener(user=user,
                                      password=password,
                                      database=database,
                                      host=host,
                                      process=process,
                                      listener=listener,
                                      args=args,
                                      domain_listener=domain_listener())
        return MetricsDomain(broker, user, password, self._domain)

    @property
    def nuomonitor_collect_queue(self):
        return self._nuomonitor_collect_queue

    def _nuomon_handler_thread(self):
        obj = NuoMonHandler(self)
        obj.start()
        self._numon_handler_ready = True
        time.sleep(10)
        self.wait_for_terminate()
        nuoca_log(logging.INFO, "NuoMon: NuoDB connection terminated")
        self._domain_metrics.disconnect()
        self._numon_handler_ready = False
        while not self._numon_handler_ready:
            time.sleep(self._collection_interval)
            nuoca_log(logging.INFO, "NuoMon: calling startup")
            self.startup_nuomon()

    def startup_nuomon(self):
        try:
            self._numon_handler_ready = False
            self._domain_metrics = \
                self.get_nuodb_metrics(
                    self._broker,
                    self._domain_password,
                    listener=MetricsProducer,
                    user=self._domain_username,
                    host=self._domain_metrics_host)
            self._thread = threading.Thread(target=self._nuomon_handler_thread)
            self._thread.daemon = True
            self._thread.start()
            try_count = 0
            while not self._numon_handler_ready and try_count < 5:
                try_count += 1
                time.sleep(1)
            return self._numon_handler_ready
        except Exception as e:
            nuoca_log(logging.ERROR, "NuoMon Startup error: %s" % str(e))
            return False

    def startup(self, config=None):
        try:
            self._config = config

            # Validate the configuration.
            required_config_items = ['broker', 'domain_username', 'domain_password']
            if not self.has_required_config_items(config, required_config_items):
                return False

            # Don't reveal the domain password in the NuoCA log file.
            display_config = {}
            display_config.update(config)
            display_config['domain_password'] = ''
            nuoca_log(logging.INFO, "NuoMon: plugin config: %s" %
                      str(display_config))

            self._broker = os.path.expandvars(config['broker'])
            self._domain_username = os.path.expandvars(config['domain_username'])
            self._domain_password = os.path.expandvars(config['domain_password'])
            if 'domain_metrics_host' in config:
                self._domain_metrics_host = os.path.expandvars(config['domain_metrics_host'])
                if self._domain_metrics_host == 'localhost':
                    self._domain_metrics_host = socket.gethostname()
            if 'database_regex_pattern' in config:
                self._database_regex_pattern = config['database_regex_pattern']
            if 'host_uuid_shortname' in config:
                self._host_uuid_shortname = config['host_uuid_shortname']
            self.startup_nuomon()
            return self._numon_handler_ready
        except Exception as e:
            nuoca_log(logging.ERROR, "NuoMon Plugin Startup error: %s" % str(e))
            return False

    def shutdown(self):
        self._enabled = False
        pass

    def collect(self, collection_interval):
        uuid_hostname_regex = \
            '^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}-'
        rval = None
        self._collection_interval = collection_interval
        try:
            nuoca_log(logging.DEBUG, "NuoMon: collect()")
            base_values = super(NuoMonitorPlugin, self).collect(collection_interval)
            collection_count = len(self._nuomonitor_collect_queue)
            nuoca_log(logging.DEBUG, "NuoMon: collection_count %s" %
                      str(collection_count))
            if not collection_count:
                return rval

            rval = []
            for i in range(collection_count):
                collected_dict = self._nuomonitor_collect_queue.pop(0)
                if self._domain_metrics_host:
                    if collected_dict['Hostname'] != self._domain_metrics_host:
                        continue
                m = re.search(self._database_regex_pattern, collected_dict['Database'])
                if m:
                    if self._host_uuid_shortname:
                        m2 = re.search(uuid_hostname_regex, collected_dict['Hostname'])
                        if m2:
                            shortid = collected_dict['Hostname'][37:]
                            if 'NodeType' in collected_dict:
                                if collected_dict['NodeType'] == 'Transaction':
                                    shortid += "(TE)"
                                elif collected_dict['NodeType'] == 'Storage':
                                    shortid += "(SM)"
                            shortid_with_pid = shortid + str(collected_dict['ProcessId'])
                            collected_dict['HostShortID'] = shortid
                            collected_dict['HostShortIDwithPID'] = shortid_with_pid
                    rval.append(collected_dict)
        except Exception as e:
            nuoca_log(logging.ERROR, str(e))
        return rval



# (C) Copyright NuoDB, Inc. 2017-2020
#
# This source code is licensed under the MIT license found in the LICENSE
# file in the root directory of this source tree.

import logging
from pynuoca.nuoca_plugin import NuocaMPInputPlugin
from pynuoca.nuoca_util import nuoca_log


class CounterPlugin(NuocaMPInputPlugin):
  def __init__(self, parent_pipe):
    super(CounterPlugin, self).__init__(parent_pipe, 'Counter')
    self._config = None
    self._count = 0
    self._increment_value = 1

  def startup(self, config=None):
    try:
      self._config = config
      if config:
        if 'increment' in config:
          self._increment_value = int(config['increment'])
      return True
    except Exception as e:
      nuoca_log(logging.ERROR, str(e))
      return False

  def shutdown(self):
    pass

  def increment(self):
    self._count += self._increment_value

  def get_count(self):
    return self._count

  def collect(self, collection_interval):
    rval = None
    try:
      nuoca_log(logging.DEBUG, "Called collect() in MPCounterPlugin process")
      collected_values = super(CounterPlugin, self).collect(collection_interval)
      self.increment()
      collected_values["counter"] = self.get_count()
      rval = []
      rval.append(collected_values)
    except Exception as e:
      nuoca_log(logging.ERROR, str(e))
    return rval



# (C) Copyright NuoDB, Inc. 2017-2020
#
# This source code is licensed under the MIT license found in the LICENSE
# file in the root directory of this source tree.

import hashlib
import json
import logging
import os
import re
import socket
import subprocess
import threading
import time

from datetime import datetime
from dateutil.parser import parse as date_parse
from pynuoca.nuoca_plugin import NuocaMPInputPlugin
from pynuoca.nuoca_util import nuoca_log
from pynuoca.nuoca_util import to_bool


# Logstash plugin
#
# Logstash plugin configuration:
#
# - Logstash:
#    description: Description for your plugin config.
#    logstashBin: full path to the logstash executable (required)
#    logstashConfig: full path to a logstash config file. (required)
#    logstashOptions: logstash command line options. (optional)
#      This logstash plug always runs logstash with --node.name and
#      --path.config command line arguments.  If you want any option
#      logstash arguments, then set, space separated in logstashOptions.
#    logstashInputFilePath: full path to an optional logstash input file.
#      (optional)
#       If this is set, then a shell environment variable
#       LOGSTASH_INPUT_FILE_PATH is set to this value.  This provides
#       a convient way to parameterize the input file for logstash
#       config file.
#    logstashSincedbPath: full path to logstash sincedb file. (optional)
#      if logstashSincedbPath is not set, and logstashInputFilePath is set,
#      then logstashSincedbPath is set to:
#          $HOME/.sincedb_<MD5 hash of logstashInputFilePath>
#      If this is set, then a shell environment variable LOGSTASH_SINCEDB_PATH
#      is set to this value.  In this way, you can parameterize the
#      file.sincedb_path in a logstash config file.
#    dropThrottledEvents: boolean (optional, default=false)
#      if the is set, then drop any events with the "throttled" tag.
#
#  Users can set 'timestamp_iso8601' in their collection to force the event
#  timestamp to become the epoch_millis (the number of milliseconds since
#  Unix Epoch)
#


class LogstashPlugin(NuocaMPInputPlugin):
    def __init__(self, parent_pipe):
        super(LogstashPlugin, self).__init__(parent_pipe, 'Logstash')
        self._config = None
        self._enabled = False
        self._logstash_bin = None
        self._logstash_config = None
        self._nuocaCollectionName = None
        self._logstash_thread = None
        self._process_thread = None
        self._logstash_subprocess = None
        self._logstash_sincedb_path = None
        self._logstash_options = None
        self._drop_throttled = False
        self._line_counter = 0
        self._lines_processed = 0
        self._local_hostname = socket.gethostname()
        self._host_uuid_shortname = False
        self._host_shortid = None
        self._logstash_collect_queue = []

    @property
    def logstash_collect_queue(self):
        return self._logstash_collect_queue

    def _run_logstash_thread(self):
        self._logstash_subprocess = None
        msg = "Env: %s=%s" % ('NUOCA_HOME', os.environ['NUOCA_HOME'])
        nuoca_log(logging.INFO, msg)

        if self._logstash_sincedb_path:
            msg = "Env: LOGSTASH_SINCEDB_PATH=%s" % \
                  self._logstash_sincedb_path
            nuoca_log(logging.INFO, msg)
            os.environ["LOGSTASH_SINCEDB_PATH"] = self._logstash_sincedb_path
        try:
            popen_args = [self._logstash_bin,
                          '--node.name', self._nuocaCollectionName,
                          "--path.config", self._logstash_config]
            if self._logstash_options:
                popen_args.extend(self._logstash_options)
            msg = "Starting %s --node.name %s --path.config %s" % \
                  (self._logstash_bin,
                   self._nuocaCollectionName,
                   self._logstash_config)
            for item in self._logstash_options:
                msg += " %s" % str(item)
            nuoca_log(logging.INFO, msg)
            self._logstash_subprocess = \
                subprocess.Popen(popen_args, stdout=subprocess.PIPE)
        except Exception as e:
            msg = "logstash process: %s" % str(e)
            msg += "\nlogstash_bin: %s" % self._logstash_bin
            msg += "\nlogstash_config: %s" % self._logstash_config
            nuoca_log(logging.ERROR, msg)
            return

        try:
            while self._enabled:
                json_object = None
                line = self._logstash_subprocess.stdout.readline()
                if not line:
                    # Check to see if the logstash process failed unexpectedly
                    if self._logstash_subprocess.poll() != None:
                        self._enabled = False
                        msg = "logstash process has unexpectedly exited, returncode:%s" % \
                              self._logstash_subprocess.returncode
                        nuoca_log(logging.ERROR, msg)
                else:
                    self._line_counter += 1
                    try:
                        json_object = json.loads(line)
                    except ValueError:
                        # These messages are typical log messages about logstash itself
                        # which are written to stdout.  Just write them to the nuoca.log
                        msg = "logstash message: %s" % line
                        nuoca_log(logging.INFO, msg)
                    if json_object:
                        self._logstash_collect_queue.append(json_object)
            nuoca_log(logging.INFO,
                      "Logstash plugin run_logstash_thread "
                      "completed %s lines" % str(self._line_counter))
        except Exception as e:
            msg = "logstash process: %s" % str(e)
            nuoca_log(logging.ERROR, msg)
            pass

        try:
            if self._logstash_subprocess:
                self._logstash_subprocess.kill()
        except Exception as e:
            msg = "Exception trying to kill logstash process: %s" % str(e)
            nuoca_log(logging.ERROR, msg)

    def startup(self, config=None):
        uuid_hostname_regex = \
            '^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}-'
        try:
            self._config = config

            # Validate the configuration.
            #   logstash_bin: full path to the logstash executable
            #   logstash_config: full path to a logstash config file.
            required_config_items = ['logstashBin', 'logstashConfig']
            if not self.has_required_config_items(config, required_config_items):
                return False
            nuoca_log(logging.INFO, "Logstash plugin config: %s" %
                      str(self._config))

            self._logstash_bin = os.path.expandvars(config['logstashBin'])
            if not os.path.isfile(self._logstash_bin):
                msg = "Unable to find 'logstashBin' file: %s" % self._logstash_bin
                nuoca_log(logging.ERROR, msg)
                return False

            self._logstash_config = os.path.expandvars(config['logstashConfig'])
            if not os.path.isfile(self._logstash_config):
                msg = "Unable to find 'logstashConfig' file: %s" % \
                      self._logstash_config
                nuoca_log(logging.ERROR, msg)
                return False
            nuoca_log(logging.INFO, "logstashConfig: %s" % str(self._logstash_config))

            logstash_input_file_path = '/dev/null'
            logstash_input_file_path2 = '/dev/null'

            if 'logstashInputFilePath' in self._config:
                if isinstance(self._config['logstashInputFilePath'], str):
                    msg = "Env: LOGSTASH_INPUT_FILE_PATH=%s" % \
                          self._config['logstashInputFilePath']
                    nuoca_log(logging.INFO, msg)
                    logstash_input_file_path = \
                        os.path.expandvars(self._config['logstashInputFilePath'])

            if isinstance(self._config['logstashInputFilePath'], list):
                if len(self._config['logstashInputFilePath']) > 2:
                    msg = "logstashInputFilePath list has more than 2 elements: %s" % \
                          self._config['logstashInputFilePath']
                    nuoca_log(logging.ERROR, msg)
                    return
                else:
                    msg = "Env: LOGSTASH_INPUT_FILE_PATH=%s" % \
                          self._config['logstashInputFilePath'][0]
                    nuoca_log(logging.INFO, msg)
                    logstash_input_file_path = \
                        os.path.expandvars(self._config['logstashInputFilePath'][0])
                    logstash_input_file_path2 = \
                        os.path.expandvars(self._config['logstashInputFilePath'][1])

            os.environ["LOGSTASH_INPUT_FILE_PATH"] = logstash_input_file_path
            msg = "Env: LOGSTASH_INPUT_FILE_PATH=%s" % logstash_input_file_path
            nuoca_log(logging.INFO, msg)
            os.environ["LOGSTASH_INPUT_FILE_PATH2"] = logstash_input_file_path2
            msg = "Env: LOGSTASH_INPUT_FILE_PATH2=%s" % logstash_input_file_path2
            nuoca_log(logging.INFO, msg)
            if 'logstashSincedbPath' in config:
                self._logstash_sincedb_path = \
                    os.path.expandvars(config['logstashSincedbPath'])
            else:
                if 'logstashInputFilePath' in config:
                    nuoca_log(logging.INFO, "Logstash Plugin Input File Path: %s" %
                              str(config['logstashInputFilePath']))
                    hexdigest = hashlib.md5(str(config['logstashInputFilePath'])).hexdigest()
                    self._logstash_sincedb_path = \
                        "%s/.sincedb_%s" % (os.environ['HOME'], hexdigest)

            nuoca_log(logging.INFO, "Logstash Plugin sincedb_path: %s" %
                      str(self._logstash_sincedb_path))

            if 'logstashOptions' in config:
                logstash_options = os.path.expandvars(config['logstashOptions'])
                self._logstash_options = logstash_options.split(' ')

            if 'dropThrottledEvents' in config:
                self._drop_throttled = to_bool(config['dropThrottledEvents'])

            if 'nuocaCollectionName' in config:
                self._nuocaCollectionName = config['nuocaCollectionName']

            # For Coach hostnames in the format: uuid-shortId
            if 'host_uuid_shortname' in config:
                self._host_uuid_shortname = config['host_uuid_shortname']

            if self._host_uuid_shortname:
                m2 = re.search(uuid_hostname_regex, self._local_hostname)
                if m2:
                    self._host_shortid = self._local_hostname[37:]

            self._enabled = True
            self._logstash_thread = \
                threading.Thread(target=self._run_logstash_thread)
            self._logstash_thread.daemon = True
            self._logstash_thread.start()

            return True

        except Exception as e:
            nuoca_log(logging.ERROR, "Logstash Plugin: %s" % str(e))
            return False

    def shutdown(self):
        self._enabled = False
        if self._logstash_subprocess:
            self._logstash_subprocess.terminate()
            self._logstash_subprocess = None
        if self._process_thread:
            self._process_thread.join()

    def collect(self, collection_interval):
        rval = None
        if not self._enabled:
            return 0
        dt1 = datetime.utcnow()
        drop_throttled_count = 0
        try:
            nuoca_log(logging.DEBUG,
                      "Called collect() in Logstash Plugin process")
            base_values = super(LogstashPlugin, self). \
                collect(collection_interval)
            base_values['Hostname'] = self._local_hostname
            if self._host_shortid:
                base_values['HostShortID'] = self._host_shortid
            if self._nuocaCollectionName:
                base_values['nuocaCollectionName'] = self._nuocaCollectionName
            rval = []
            collection_count = len(self._logstash_collect_queue)
            if not collection_count:
                return rval

            for i in range(collection_count):
                collected_dict = self._logstash_collect_queue.pop(0)
                collected_dict.update(base_values)
                if self._drop_throttled and \
                        'tags' in collected_dict and \
                        'throttled' in collected_dict['tags']:
                    drop_throttled_count += 1
                    continue
                if 'timestamp_iso8601' in collected_dict:
                    dt1 = date_parse(collected_dict['timestamp_iso8601'])
                    tt = dt1.timetuple()
                    # When processing the ISO8601 time string, there is no way to
                    # determine local daylight saving time from the timestring, so the
                    # call to dt1.timetuple() sets tm_idst=0.  But we want to force
                    # time.mktime() to consider daylight savings time in it's calculation
                    # so we need to set tm_isdst=-1.  The Python time tuple 'tt' is
                    # read-only, so I made a writeable list 'tt_writeable' to set
                    # tm_isdst (the 9th element in the list) to -1.  And then create
                    # a new time tuple 'tt' from the writeable version.
                    tt_writeable = list(tt)
                    tt_writeable[8] = -1
                    tt = time.struct_time(tuple(tt_writeable))
                    epoch_seconds = int(time.mktime(tt))
                    epoch_millis = epoch_seconds * 1000 + dt1.microsecond / 1000
                    collected_dict['TimeStamp'] = epoch_millis
                rval.append(collected_dict)

            if self._drop_throttled and drop_throttled_count > 0:
                msg = "Logstash (%s) dropping %d throttled events" % \
                      (self._nuocaCollectionName, drop_throttled_count)
                nuoca_log(logging.INFO, msg)
                drop_throttle_event = {}
                epoch_seconds = int(time.mktime(dt1.timetuple()))
                epoch_millis = epoch_seconds * 1000 + dt1.microsecond / 1000
                drop_throttle_event['TimeStamp'] = epoch_millis
                drop_throttle_event['logger'] = 'NuoCA'
                drop_throttle_event['loglevel'] = 'INFO'
                drop_throttle_event['message'] = msg
                rval.append(drop_throttle_event)

        except Exception as e:
            nuoca_log(logging.ERROR, str(e))
        return rval



# (C) Copyright NuoDB, Inc. 2017-2020
#
# This source code is licensed under the MIT license found in the LICENSE
# file in the root directory of this source tree.

import os
import yaml


class NuocaConfig(object):
    """
    NuoCA Configuration elements.
    """
    NUOCA_TMPDIR = '/tmp/nuoca'  # Temporary directory for NuoCA
    NUOCA_LOGFILE = '/tmp/nuoca/nuoca.log'  # Path to logfile for NuoCA
    PLUGIN_PIPE_TIMEOUT = 5  # Plugin communication pipe timeout in seconds
    NUOCA_CONFIG_FILE = None
    SELFTEST_LOOP_COUNT = 5  # Number of Collection Intervals in selftest.
    SUBPROCESS_EXIT_TIMEOUT = 5  # Max seconds to wait for subprocess exit
    NUOCA_COLLECTION_INTERVAL = None  # Override NuoCA Collection Interval

    # Plugins that will be populated from the NuoCA configuration file.
    INPUT_PLUGINS = []
    OUTPUT_PLUGINS = []
    TRANSFORM_PLUGINS = []

    def _validate(self, userconfig):
        # TODO Implement
        pass

    # Check each Input plugin configuration for 'NuoCA.route'. If found
    # check it against each Output plugin, and remove any routes
    # that don't match an actual Output plugin.  Essentially rewriting
    # each 'NuoCA.route' configuration to match actual routes.
    def _adjust_route(self, userconfig):
        if not userconfig:
            return
        if 'INPUT_PLUGINS' in userconfig:
            for input_plugin in userconfig['INPUT_PLUGINS']:
                input_plugin_name = input_plugin.keys()[0]
                input_plugin_config = input_plugin.values()[0]
                try:
                    if 'NuoCA.route' in input_plugin_config:
                        self.is_routing = True
                        actual_route_list = []
                        candidate_route_list = \
                            input_plugin_config['NuoCA.route'].split(',')
                        for candidate_route in candidate_route_list:
                            for output_plugin in userconfig['OUTPUT_PLUGINS']:
                                output_plugin_name = output_plugin.keys()[0]
                                if candidate_route == output_plugin_name:
                                    actual_route_list.append(output_plugin_name)
                        input_plugin_config['NuoCA.route'] = ','.join(actual_route_list)
                except:
                    pass

    def __init__(self, config_file):
        self.is_routing = False
        if not config_file:
            raise AttributeError("You must provide a NuoCA Config file")
        if not os.path.exists(config_file):
            raise AttributeError("Config file: %s does not exist" % config_file)
        userconfig = yaml.safe_load(open(config_file).read())
        self._validate(userconfig)
        self._adjust_route(userconfig)
        if not userconfig:
            return
        self.NUOCA_CONFIG_FILE = config_file
        for key, value in userconfig.iteritems():
            setattr(self, key, value)
